import contextlib
import dataclasses
import enum
import io
import json
import os
import pathlib
import sys
from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar, Union, cast

import pytest
from helptext_utils import get_helptext_with_checks
from typing_extensions import Annotated, Literal, NotRequired, TypedDict

import tyro
import tyro._strings


def test_helptext() -> None:
    @dataclasses.dataclass
    class Helptext:
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: Annotated[int, "ignored"]

        z: int = 3
        """Documentation 3"""

    helptext = get_helptext_with_checks(Helptext)
    assert cast(str, helptext) in helptext
    assert "x INT" in helptext
    assert "y INT" in helptext
    assert "z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_helptext_sphinx_autodoc_style() -> None:
    @dataclasses.dataclass
    class Helptext:
        """This docstring should be printed as a description."""

        x: int  #: Documentation 1

        #:Documentation 2
        y: Annotated[int, "ignored"]
        z: int = 3

    helptext = get_helptext_with_checks(Helptext)
    assert cast(str, helptext) in helptext
    assert "x INT" in helptext
    assert "y INT" in helptext
    assert "z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert ": Documentation 1" not in helptext
    assert "Documentation 2 (required)" in helptext
    assert ":Documentation 2" not in helptext

    # :Documentation 2 should not be applied to `z`.
    assert helptext.count("Documentation 2") == 1


def test_helptext_from_class_docstring() -> None:
    @dataclasses.dataclass
    class Helptext2:
        """This docstring should be printed as a description.

        Attributes:
            x: Documentation 1
            y: Documentation 2
            z: Documentation 3
        """

        x: int
        y: Annotated[int, "ignored"]
        z: int = 3

    helptext = get_helptext_with_checks(Helptext2)
    assert "This docstring should be printed as a description" in helptext
    assert "Attributes" not in helptext
    assert "x INT" in helptext
    assert "y INT" in helptext
    assert "z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_helptext_from_class_docstring_args() -> None:
    @dataclasses.dataclass
    class Helptext3:
        """This docstring should be printed as a description.

        Args:
            x: Documentation 1
            y: Documentation 2
            z: Documentation 3
        """

        x: int
        y: Annotated[int, "ignored"]
        z: int = 3

    helptext = get_helptext_with_checks(Helptext3)
    assert "This docstring should be printed as a description" in helptext
    assert "Args" not in helptext
    assert "x INT" in helptext
    assert "y INT" in helptext
    assert "z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_helptext_inherited() -> None:
    class UnrelatedParentClass:
        pass

    @dataclasses.dataclass
    class ActualParentClass:
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        # fmt: off

        z: int = 3
        def some_method(self) -> None:  # noqa
            """Coverage stress test."""
        # fmt: on

    @dataclasses.dataclass
    class ChildClass(UnrelatedParentClass, ActualParentClass):
        pass

    helptext = get_helptext_with_checks(ChildClass)
    assert "Documentation 1" in helptext
    assert "Documentation 2" in helptext


def test_helptext_inherited_default_override() -> None:
    @dataclasses.dataclass
    class ParentClass:
        """This docstring should __not__ be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        # fmt: off

        z: int = 3
        def some_method(self) -> None:  # noqa
            """Coverage stress test."""
        # fmt: on

    @dataclasses.dataclass
    class ChildClass(ParentClass):
        """This docstring should be printed as a description."""

    def main(x: ParentClass = ChildClass(x=5, y=5)) -> Any:
        """Main function."""
        return x

    helptext = get_helptext_with_checks(main)
    assert "Main function." in helptext
    assert "Documentation 1" in helptext
    assert "Documentation 2" in helptext
    assert "__not__" not in helptext
    assert helptext.count("should be printed") == 1


def test_helptext_nested() -> None:
    """For nested classes, we should pull helptext from the outermost docstring if
    possible. The class docstring can be used as a fallback."""

    class Inner:
        def __init__(self, a: int):
            """Something

            Args:
                a (int): Hello world!
            """

    def main_with_docstring(a: Inner) -> None:
        """main_with_docstring.

        Args:
            a: Documented in function.
        """

    def main_no_docstring(a: Inner) -> None:
        """main_no_docstring."""

    helptext = get_helptext_with_checks(main_with_docstring)
    assert "Documented in function" in helptext and str(Inner.__doc__) not in helptext
    assert "main_with_docstring." in helptext
    assert "Args:" not in helptext
    assert "Args:" not in helptext
    assert "Hello world!" in helptext

    helptext = get_helptext_with_checks(main_no_docstring)
    assert "Something" in helptext
    assert "main_no_docstring." in helptext
    assert "Args:" not in helptext
    assert "Hello world!" in helptext


@pytest.mark.skipif(sys.platform == "win32", reason="Path separators differ on Windows")
def test_helptext_defaults() -> None:
    class Color(enum.Enum):
        RED = enum.auto()
        GREEN = enum.auto()
        BLUE = enum.auto()

    @dataclasses.dataclass
    class HelptextWithVariousDefaults:
        x: pathlib.Path = pathlib.Path("/some/path/to/a/file")
        y: Color = Color.RED
        z: str = "%"

    helptext = get_helptext_with_checks(HelptextWithVariousDefaults)
    assert "show this help message and exit" in helptext
    assert "--x PATH" in helptext
    assert "(default: /some/path/to/a/file)" in helptext
    assert "--y {RED,GREEN,BLUE}" in helptext
    assert "(default: RED)" in helptext
    assert "--z STR" in helptext
    assert "(default: %)" in helptext


def test_helptext_upath_defaults() -> None:
    """Test UPath helptext with metavar and default rendering."""
    from upath import UPath

    @dataclasses.dataclass
    class HelptextWithUPath:
        x: UPath = UPath("s3://bucket/path")

    helptext = get_helptext_with_checks(HelptextWithUPath)
    assert "--x UPATH" in helptext
    assert "(default: s3://bucket/path)" in helptext


def test_multiline_helptext() -> None:
    @dataclasses.dataclass
    class HelptextMultiline:
        x: int  # Documentation 1

        # This comment should be ignored!

        # Documentation 2
        # Next line of documentation 2
        y: int

        z: int = 3
        """Documentation 3
        Next line of documentation 3"""

    helptext = get_helptext_with_checks(HelptextMultiline)
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2" in helptext
    assert "documentation 2" in helptext
    assert "Documentation 3" in helptext
    assert "documentation 3" in helptext


def test_grouped_helptext() -> None:
    @dataclasses.dataclass
    class HelptextGrouped:
        x: int  # Documentation 1
        # Description of both y and z.
        y: int
        z: int = 3

    helptext = get_helptext_with_checks(HelptextGrouped)
    assert "Documentation 1 (required)" in helptext
    assert "Description of both y and z. (required)" in helptext
    assert "Description of both y and z. (default: 3)" in helptext


def test_none_default_value_helptext() -> None:
    @dataclasses.dataclass
    class Config:
        x: Optional[int] = None
        """An optional variable."""

    helptext = get_helptext_with_checks(Config)
    assert "--x {None}|INT" in helptext
    assert "An optional variable. (default: None)" in helptext


def test_helptext_hard_bool() -> None:
    @dataclasses.dataclass
    class HelptextHardString:
        # fmt: off
        x: bool = (
            False
        )
        """Helptext. 2% milk."""
        # fmt: on

    # The percent symbol needs some extra handling in argparse.
    # https://stackoverflow.com/questions/21168120/python-argparse-errors-with-in-help-string

    helptext = get_helptext_with_checks(HelptextHardString)
    assert "--x" in helptext
    assert "2% milk." in helptext


def test_helptext_with_inheritance() -> None:
    @dataclasses.dataclass
    class Parent:
        # fmt: off
        x: str = (
            "This docstring may be tougher to parse!"
        )
        """Helptext."""
        # fmt: on

    @dataclasses.dataclass
    class Child(Parent):
        pass

    helptext = get_helptext_with_checks(Child)
    assert "--x STR" in helptext
    assert "Helptext." in helptext
    assert "(default: 'This docstring" in helptext


def test_helptext_with_inheritance_overriden() -> None:
    @dataclasses.dataclass
    class Parent2:
        # fmt: off
        x: str = (
            "This docstring may be tougher to parse!"
        )
        """Helptext."""
        # fmt: on

    @dataclasses.dataclass
    class Child2(Parent2):
        # fmt: off
        x: str = (
            "This docstring may be tougher to parse?"
        )
        """Helptext!"""
        # fmt: on

    helptext = get_helptext_with_checks(Child2)
    assert "--x STR" in helptext
    assert "Helptext! (default: 'This" in helptext


def test_tuple_helptext() -> None:
    @dataclasses.dataclass
    class TupleHelptext:
        x: Tuple[int, str, float]

    helptext = get_helptext_with_checks(TupleHelptext)
    assert "--x INT STR FLOAT" in helptext


def test_tuple_helptext_defaults() -> None:
    @dataclasses.dataclass
    class TupleHelptextDefaults:
        x: Tuple[int, str, str] = (5, "hello world", "hello")

    helptext = get_helptext_with_checks(TupleHelptextDefaults)
    assert "--x INT STR STR" in helptext
    assert "(default: 5 'hello world' hello)" in helptext


def test_generic_helptext() -> None:
    T = TypeVar("T")

    @dataclasses.dataclass
    class GenericTupleHelptext(Generic[T]):
        x: T

    helptext = get_helptext_with_checks(GenericTupleHelptext[int])
    assert "--x INT" in helptext


def test_generic_tuple_helptext() -> None:
    T = TypeVar("T")

    @dataclasses.dataclass
    class GenericTupleHelptext(Generic[T]):
        x: Tuple[T, T, T]

    helptext = get_helptext_with_checks(GenericTupleHelptext[int])
    assert "--x INT INT INT" in helptext


def test_generic_list_helptext() -> None:
    T = TypeVar("T")

    @dataclasses.dataclass
    class GenericTupleHelptext(Generic[T]):
        x: List[T]

    helptext = get_helptext_with_checks(GenericTupleHelptext[int])
    assert "--x [INT [INT ...]]" in helptext


def test_literal_helptext() -> None:
    @dataclasses.dataclass
    class LiteralHelptext:
        x: Literal[1, 2, 3]
        """A number."""

    helptext = get_helptext_with_checks(LiteralHelptext)
    assert "--x {1,2,3}" in helptext
    assert "A number. (required)" in helptext


def test_optional_literal_helptext() -> None:
    @dataclasses.dataclass
    class OptionalLiteralHelptext:
        x: Optional[Literal[1, 2, 3]] = None
        """A number."""

    helptext = get_helptext_with_checks(OptionalLiteralHelptext)
    assert "--x {None,1,2,3}" in helptext
    assert "A number. (default: None)" in helptext


def test_multiple_subparsers_helptext() -> None:
    @dataclasses.dataclass
    class Subcommand1:
        """2% milk."""  # % symbol is prone to bugs in argparse.

        x: int = 0

    @dataclasses.dataclass
    class Subcommand2:
        y: int = 1

    @dataclasses.dataclass
    class Subcommand3:
        z: int = 2

    @dataclasses.dataclass
    class MultipleSubparsers:
        # Field a description.
        a: Union[Subcommand1, Subcommand2, Subcommand3]
        # Field b description.
        b: Union[Subcommand1, Subcommand2, Subcommand3]
        # Field c description.
        c: Union[Subcommand1, Subcommand2, Subcommand3] = dataclasses.field(
            default_factory=Subcommand3
        )

        d: bool = False

    helptext = get_helptext_with_checks(MultipleSubparsers)

    assert "2% milk." in helptext
    assert "Field a description." in helptext
    # With tyro backend, all subparser groups in the frontier are shown.
    # With argparse backend, only the first subparser group is shown.

    if tyro._experimental_options["backend"] == "tyro":
        assert "Field b description." in helptext
        assert "Field c description." in helptext
    else:
        assert "Field b description." not in helptext
        assert "Field c description." not in helptext

    # Not enough args for usage shortening to kick in.
    assert "[OPTIONS]" not in helptext
    assert "[B:SUBCOMMAND2 OPTIONS]" not in helptext

    helptext = get_helptext_with_checks(
        MultipleSubparsers,
        args=["a:subcommand1", "b:subcommand1", "--help"],
    )

    assert "2% milk." in helptext
    assert "Field a description." not in helptext
    assert "Field b description." not in helptext
    assert "Field c description." in helptext
    assert "--no-d" not in helptext
    assert "(default: c:subcommand3)" in helptext

    # Not enough args for usage shortening to kick in.
    assert "[OPTIONS]" not in helptext
    assert "[B:SUBCOMMAND1 OPTIONS]" not in helptext
    assert "[B:SUBCOMMAND2 OPTIONS]" not in helptext

    # Argument should be pushed to the leaf.
    assert "--d, --no-d" in get_helptext_with_checks(
        MultipleSubparsers,
        args=["a:subcommand1", "b:subcommand1", "c:subcommand2", "--help"],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )


def test_multiple_subparsers_helptext_shortened_usage() -> None:
    @dataclasses.dataclass
    class Subcommand1:
        """2% milk."""  # % symbol is prone to bugs in argparse.

        aaaaaaa: int = 0
        bbbbbbb: int = 0
        ccccccc: int = 0
        ddddddd: int = 0
        eeeeeee: int = 0

    @dataclasses.dataclass
    class Subcommand2:
        aaaaaaa: int = 0
        bbbbbbb: int = 0
        ccccccc: int = 0
        ddddddd: int = 0
        eeeeeee: int = 0

    @dataclasses.dataclass
    class Subcommand3:
        a: int = 0
        b: int = 0
        c: int = 0
        d: int = 0
        e: int = 0

    @dataclasses.dataclass
    class MultipleSubparsers:
        # Field a description.
        a: Union[Subcommand1, Subcommand2, Subcommand3]
        # Field b description.
        b: Union[Subcommand1, Subcommand2, Subcommand3]
        # Field c description.
        c: Union[Subcommand1, Subcommand2, Subcommand3] = dataclasses.field(
            default_factory=Subcommand3
        )

        ddddddd: bool = False
        fffffff: bool = False
        ggggggg: bool = False
        hhhhhhh: bool = False
        iiiiiii: bool = False

    helptext = get_helptext_with_checks(MultipleSubparsers)

    assert "2% milk." in helptext
    assert "Field a description." in helptext
    # With tyro backend, all subparser groups in the frontier are shown.
    # With argparse backend, only the first subparser group is shown.

    if tyro._experimental_options["backend"] == "tyro":
        assert "Field b description." in helptext
        assert "Field c description." in helptext
    else:
        assert "Field b description." not in helptext
        assert "Field c description." not in helptext

    assert "[OPTIONS]" in helptext
    assert "[B:SUBCOMMAND2 OPTIONS]" not in helptext

    helptext = get_helptext_with_checks(
        MultipleSubparsers, args=["a:subcommand1", "b:subcommand1", "--help"]
    )

    assert "2% milk." in helptext
    assert "Field a description." not in helptext
    assert "Field b description." not in helptext
    assert "Field c description." in helptext
    assert "(default: c:subcommand3)" in helptext

    assert "[OPTIONS]" not in helptext
    assert "[B:SUBCOMMAND1 OPTIONS]" in helptext
    assert "[B:SUBCOMMAND2 OPTIONS]" not in helptext


def test_optional_helptext() -> None:
    @dataclasses.dataclass
    class OptionalHelptext:
        """This docstring should be printed as a description. 2% milk."""

        x: Optional[int]  # Documentation 1

        # Documentation 2
        y: List[Optional[int]]

        z: Optional[int] = 3
        """Documentation 3"""

    helptext = get_helptext_with_checks(OptionalHelptext)
    assert cast(str, cast(str, OptionalHelptext.__doc__)[:20]) in helptext
    assert "2% milk" in helptext
    assert "--x {None}|INT" in helptext
    assert "--y [{None}|INT [{None}|INT ...]]" in helptext
    assert "[--z {None}|INT]" in helptext


def test_metavar_0() -> None:
    def main(x: Union[Literal[0, 1, 2, 3], Tuple[int, int]]) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3}|{INT INT}" in helptext


def test_metavar_1() -> None:
    def main(
        x: Union[
            Literal[0, 1, 2, 3],
            Literal["hey,there", "hello"],
            List[int],
        ],
    ) -> None:
        pass

    # The comma formatting is unfortunate, but matches argparse's default behavior.
    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3,hey,there,hello}|{[INT [INT ...]]}" in helptext


def test_metavar_2() -> None:
    def main(
        x: Tuple[
            Literal[0, 1, 2, 3],
            Union[int, str],
        ],
    ) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3} INT|STR" in helptext


def test_metavar_3() -> None:
    def main(
        x: Union[
            Literal[0, 1, 2, 3],
            Union[Tuple[int, int], Tuple[str]],
        ],
    ) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3}|{INT INT}|STR" in helptext


def test_metavar_4() -> None:
    def main(
        x: Union[
            Literal[0, 1, 2, 3],
            Union[Tuple[int, int], Tuple[str, str, str]],
            Literal[True],
        ],
    ) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3}|{INT INT}|{STR STR STR}|{True}" in helptext


def test_metavar_5() -> None:
    def main(
        x: List[Union[Tuple[int, int], Tuple[str, str]]] = [(1, 1), (2, 2)],
    ) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "[--x [{INT INT}|{STR STR} [{INT INT}|{STR STR} ...]]]" in helptext


def test_metavar_6() -> None:
    def main(x: Dict[Union[Tuple[int, int], Tuple[str, str]], Tuple[int, int]]) -> dict:
        return x

    helptext = get_helptext_with_checks(main)
    assert (
        "--x [{INT INT}|{STR STR} INT INT [{INT INT}|{STR STR} INT INT ...]]"
        in helptext
    )


def test_empty_metavar() -> None:
    """Test that empty metavar doesn't add trailing space."""

    @dataclasses.dataclass
    class Args:
        verbose: Annotated[int, tyro.conf.arg(metavar="", aliases=("-v",))] = 0

    helptext = get_helptext_with_checks(Args)
    # Should have no space after comma: "-v, --verbose".
    # Not: "-v , --verbose".
    assert "-v, --verbose" in helptext
    assert "-v , --verbose" not in helptext


def test_comment_in_subclass_list() -> None:
    @dataclasses.dataclass
    class Something(
        # This text should not show up in the helptext anywhere.
        object,
    ):
        a: int

        # But this text should!
        b: int

    helptext = get_helptext_with_checks(Something)
    assert "This text should not" not in helptext
    assert "But this text should!" in helptext


def test_pathlike() -> None:
    def main(x: os.PathLike) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x PATHLIKE " in helptext


def test_nested_bool() -> None:
    @dataclasses.dataclass
    class Child:
        x: bool = False

    def main(child: Child) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--child.x | --child.no-x" in helptext


def test_multiple_subparsers_helptext_hyphens() -> None:
    @dataclasses.dataclass
    class SubcommandOne:
        """2% milk."""  # % symbol is prone to bugs in argparse.

        arg_x: int = 0
        arg_flag: bool = False

    @dataclasses.dataclass
    class SubcommandTwo:
        arg_y: int = 1

    @dataclasses.dataclass
    class SubcommandThree:
        arg_z: int = 2

    @dataclasses.dataclass
    class MultipleSubparsers:
        # Field a description.
        a: Union[SubcommandOne, SubcommandTwo, SubcommandThree]
        # Field b description.
        b: Union[SubcommandOne, SubcommandTwo, SubcommandThree]
        # Field c description.
        c: Union[SubcommandOne, SubcommandTwo, SubcommandThree] = dataclasses.field(
            default_factory=SubcommandThree
        )

    helptext = get_helptext_with_checks(MultipleSubparsers)

    assert "2% milk." in helptext
    assert "Field a description." in helptext
    # With tyro backend, all subparser groups in the frontier are shown.
    # With argparse backend, only the first subparser group is shown.

    if tyro._experimental_options["backend"] == "tyro":
        assert "Field b description." in helptext
        assert "Field c description." in helptext
    else:
        assert "Field b description." not in helptext
        assert "Field c description." not in helptext

    helptext = get_helptext_with_checks(
        MultipleSubparsers, args=["a:subcommand-one", "b:subcommand-one", "--help"]
    )

    assert "2% milk." in helptext
    assert "Field a description." not in helptext
    assert "Field b description." not in helptext
    assert "Field c description." in helptext
    assert "(default: c:subcommand-three)" in helptext
    assert "--b.arg-x" in helptext
    assert "--b.no-arg-flag" in helptext
    assert "--b.arg-flag" in helptext


def test_multiple_subparsers_helptext_underscores() -> None:
    @dataclasses.dataclass
    class SubcommandOne:
        """2% milk."""  # % symbol is prone to bugs in argparse.

        arg_x: int = 0
        arg_flag: bool = False

    @dataclasses.dataclass
    class SubcommandTwo:
        arg_y: int = 1

    @dataclasses.dataclass
    class SubcommandThree:
        arg_z: int = 2

    @dataclasses.dataclass
    class MultipleSubparsers:
        # Field a description.
        a: Union[SubcommandOne, SubcommandTwo, SubcommandThree]
        # Field b description.
        b: Union[SubcommandOne, SubcommandTwo, SubcommandThree]
        # Field c description.
        c: Union[SubcommandOne, SubcommandTwo, SubcommandThree] = dataclasses.field(
            default_factory=SubcommandThree
        )

    helptext = get_helptext_with_checks(MultipleSubparsers, use_underscores=True)

    assert "2% milk." in helptext
    assert "Field a description." in helptext
    # With tyro backend, all subparser groups in the frontier are shown.
    # With argparse backend, only the first subparser group is shown.

    if tyro._experimental_options["backend"] == "tyro":
        assert "Field b description." in helptext
        assert "Field c description." in helptext
    else:
        assert "Field b description." not in helptext
        assert "Field c description." not in helptext

    helptext = get_helptext_with_checks(
        MultipleSubparsers,
        args=["a:subcommand_one", "b:subcommand_one", "--help"],
        use_underscores=True,
    )

    assert "2% milk." in helptext
    assert "Field a description." not in helptext
    assert "Field b description." not in helptext
    assert "Field c description." in helptext
    assert "(default: c:subcommand_three)" in helptext
    assert "--b.arg_x" in helptext
    assert "--b.no_arg_flag" in helptext
    assert "--b.arg_flag" in helptext


def test_subparsers_wrapping() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: int

    @dataclasses.dataclass
    class CheckoutCompletion:
        """Help message."""

        y: int

    help = get_helptext_with_checks(Union[A, CheckoutCompletion])  # type: ignore
    # Both backends use full metavar when there's a single subparser group.
    # Appears 3 times: usage line + subcommand list.
    assert help.count("checkout-completion") == 2


def test_subparsers_wrapping1() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: int

    @dataclasses.dataclass
    class CheckoutCompletio:
        """Help message."""

        y: int

    help = get_helptext_with_checks(Union[A, CheckoutCompletio])  # type: ignore
    # Both backends use full metavar when there's a single subparser group.
    assert help.count("checkout-completio") == 2


def test_subparsers_wrapping2() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: int

    @dataclasses.dataclass
    class CheckoutCompletionn:
        """Help message."""

        y: int

    help = get_helptext_with_checks(Union[A, CheckoutCompletionn])  # type: ignore
    # Both backends use full metavar when there's a single subparser group.
    assert help.count("checkout-completionn") == 2


def test_subparsers_wrapping3() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: int

    @dataclasses.dataclass
    class CmdCheckout012:
        """Help message."""

        y: int

    help = get_helptext_with_checks(Union[A, CmdCheckout012])  # type: ignore
    # Both backends use full metavar when there's a single subparser group.
    assert help.count("cmd-checkout012") == 2


def test_tuple_default() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: Tuple[str, str] = ("hello", "world")

    help = get_helptext_with_checks(A)
    assert "STR STR" in help
    assert "hello world" in help
    assert "('hello', 'world')" not in help


def test_argconf_constructor() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: Annotated[
            Tuple[str, str], tyro.conf.arg(constructor=lambda x: ("a", "b"))
        ] = ("hello", "world")

    help = get_helptext_with_checks(A)
    assert "STR STR" not in help
    # Unlike case above, should not be converted to 'hello world'.
    assert "('hello', 'world')" in help  # JSON special case.


def test_argconf_constructor_json_special_case() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: Annotated[
            # NOTE: this style of JSON construction is/will be deprecated.
            # Avoid.
            Tuple[str, str],
            tyro.conf.arg(constructor=json.loads, metavar="JSON"),
        ] = ("hello", "world")
        y: Annotated[
            # NOTE: this style of JSON construction is/will be deprecated.
            # Avoid.
            Tuple[int, int],
            tyro.conf.arg(constructor=json.loads, metavar="JSON"),
        ] = (3, 5)

    help = get_helptext_with_checks(A)
    assert "STR STR" not in help
    assert "JSON" in help
    assert '["hello", "world"]' in help  # JSON special case.
    assert "[3, 5]" in help, help  # JSON special case.


def test_optional_group() -> None:
    def f(
        x: Annotated[
            # NOTE: this style of JSON construction is/will be deprecated.
            # Avoid.
            Tuple[str, str],
            tyro.conf.arg(constructor=json.loads, metavar="JSON"),
        ] = ("hello", "world"),
        y: int = 3,
    ) -> int:
        del x, y
        return 5

    help = get_helptext_with_checks(f, default=3)
    assert 'default if used: ["hello", "world"]' in help
    assert "default if used: 3" in help


def test_append_fixed() -> None:
    def f(
        x: tyro.conf.UseAppendAction[Tuple[str, str]],
        y: int = 3,
    ) -> int:
        del x, y
        return 5

    help = get_helptext_with_checks(f)
    assert "repeatable" not in help, help


def test_append_good() -> None:
    def f(
        x: tyro.conf.UseAppendAction[Tuple[str, ...]],
        y: int = 3,
    ) -> int:
        del x, y
        return 5

    help = get_helptext_with_checks(f)
    assert "repeatable" in help, help


def test_append_with_default() -> None:
    def f(
        x: tyro.conf.UseAppendAction[Tuple[str, ...]] = ("hello world", "hello"),
        y: int = 3,
    ) -> int:
        del x, y
        return 5

    help = get_helptext_with_checks(f)
    assert "repeatable, appends to: 'hello world' hello" in help, help


def test_typeddict_exclude() -> None:
    class Special(TypedDict):
        x: NotRequired[int]

    help = get_helptext_with_checks(Special)
    assert "unset by default" in help, help


def test_conf_erased_argname() -> None:
    @dataclasses.dataclass(frozen=True)
    class Verbose:
        is_verbose: bool = True

    @dataclasses.dataclass(frozen=True)
    class Args:
        verbose: Annotated[Verbose, tyro.conf.arg(name="")]

    def main(args: Args) -> None:
        print(args)

    helptext = get_helptext_with_checks(main)
    assert "args options" in helptext
    assert "args.verbose options" not in helptext


def test_long_literal_options() -> None:
    """Test that long literal options don't get truncated in helptext."""

    def main(
        x: Literal[
            "one",
            "two",
            "three",
            "four",
            "five",
            "six",
            "seven",
            "eight",
            "nine",
            "ten",
            "eleven",
            "twelve",
            "thirteen",
            "fourteen",
            "fifteen",
            "sixteen",
            "seventeen",
            "eighteen",
            "nineteen",
            "twenty",
        ],
    ) -> None:
        print(x)

    helptext = get_helptext_with_checks(main)

    # Check that the helptext contains all options without truncation
    for option in [
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
        "nine",
        "ten",
        "eleven",
        "twelve",
        "thirteen",
        "fourteen",
        "fifteen",
        "sixteen",
        "seventeen",
        "eighteen",
        "nineteen",
        "twenty",
    ]:
        assert option in helptext

    # Check that there's no ellipsis character in the helptext
    assert "…" not in helptext


def test_bool_help_edge_cases() -> None:
    """Test edge cases in metavar generation for unions with tuples and literals."""

    # Use the exact same pattern as in bool_help_edge.py but as individual function definitions
    # Currently this passes, but represents edge cases that were problematic
    def main_int(x_int: Union[Tuple[int], Tuple[int, int]] = (1,)) -> None:
        pass

    def main_literal(
        x_literal: Union[
            Tuple[Literal[0, 1, 2]], Tuple[Literal[0, 1, 2], Literal[0, 1, 2]]
        ] = (0,),
    ) -> None:
        pass

    def main_bool(x_bool: Union[Tuple[bool], Tuple[bool, bool]] = (True,)) -> None:
        pass

    # Test int case - should show INT|{INT INT}
    helptext_int = get_helptext_with_checks(main_int)
    assert "--x-int INT|{INT INT}" in helptext_int

    # Test literal case - should show {0,1,2}|{{0,1,2} {0,1,2}}
    helptext_literal = get_helptext_with_checks(main_literal)
    assert "--x-literal {0,1,2}|{{0,1,2} {0,1,2}}" in helptext_literal

    # Test bool case - should show {True,False}|{{True,False} {True,False}}
    helptext_bool = get_helptext_with_checks(main_bool)
    assert "--x-bool {True,False}|{{True,False} {True,False}}" in helptext_bool


def test_set_accent_color() -> None:
    """Test that set_accent_color runs without error."""
    tyro.extras.set_accent_color("blue")


def test_help_with_required_subcommands_consolidated() -> None:
    """Test that --help shows help text for required subcommands in consolidated mode.

    This is a regression test for an issue where --help would raise a 'Missing subcommand'
    error instead of showing the help text when using CascadeSubcommandArgs with
    required subcommands (no default).
    """
    import pytest

    @dataclasses.dataclass
    class ConfigA:
        a: int = 1

    @dataclasses.dataclass
    class ConfigB:
        b: int = 2

    def main(config: Union[ConfigA, ConfigB]) -> None:
        """Test function with required union subcommand.

        Args:
            config: Configuration to use.
        """
        pass

    # This should show help text and exit, not raise a "Missing subcommand" error.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--help"], config=(tyro.conf.CascadeSubcommandArgs,))


def test_implicit_arguments_show_required() -> None:
    """Test that default subcommand options show '(required)' indicator for required args."""

    @dataclasses.dataclass
    class SubcommandWithRequiredArg:
        """Subcommand with a required argument."""

        required_arg: int  # Required: no default value.
        optional_arg: str = "default"  # Optional: has default value.

    @dataclasses.dataclass
    class SubcommandAllOptional:
        """Another subcommand with all optional args."""

        x: int = 0

    def main(
        subcommand: Union[SubcommandWithRequiredArg, SubcommandAllOptional] = (
            SubcommandWithRequiredArg(required_arg=tyro.MISSING)  # type: ignore
        ),
    ) -> None:
        """Test function with default subcommand containing required arg.

        Args:
            subcommand: Subcommand choice with default.
        """
        pass

    # Get help text with CascadeSubcommandArgs enabled.
    # By default, the default subcommand's arguments appear with full descriptions.
    helptext = get_helptext_with_checks(
        main,
        args=["--help"],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )

    # Implicit arguments only appear with tyro backend.
    if tyro._experimental_options["backend"] != "tyro":
        return

    # Check that required arg has "(required)" indicator.
    assert "--subcommand.required-arg" in helptext
    assert "--subcommand.optional-arg" in helptext
    assert "(required)" in helptext


def test_multiple_default_subcommands_with_required() -> None:
    """Test multiple default subcommands showing required indicators.

    This mirrors the pattern from examples/03_subcommands/03_multiple_subcommands.py
    where dataset and optimizer are both default subcommands with required fields.
    """

    @dataclasses.dataclass
    class Mnist:
        """MNIST dataset configuration."""

        binary: bool  # Required: no default value, like the example.

    @dataclasses.dataclass
    class ImageNet:
        """ImageNet dataset configuration."""

        subset: int = 1000

    @dataclasses.dataclass
    class Adam:
        """Adam optimizer configuration."""

        learning_rate: float = 1e-3

    @dataclasses.dataclass
    class Sgd:
        """SGD optimizer configuration."""

        learning_rate: float = 3e-4

    def train(
        dataset: Union[Mnist, ImageNet] = Mnist(tyro.MISSING),  # type: ignore
        optimizer: Union[Adam, Sgd] = Adam(),
    ) -> None:
        """Example training script.

        Args:
            dataset: Dataset to train on.
            optimizer: Optimizer to train with.
        """
        pass

    # Get help text with CascadeSubcommandArgs enabled.
    helptext = get_helptext_with_checks(
        train,
        args=["--help"],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )

    # Implicit arguments only appear with tyro backend.
    if tyro._experimental_options["backend"] != "tyro":
        return

    # Check that default subcommand args are shown with full descriptions.
    assert "--dataset.binary" in helptext
    assert "(required)" in helptext
    assert "--optimizer.learning-rate" in helptext

    # Check that default subcommand groups are labeled.
    assert "source subcommand: dataset:mnist" in helptext
    assert "source subcommand: optimizer:adam" in helptext


def test_compact_help_short_text() -> None:
    """Test compact mode with short argument text that fits on one line."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class ShortArgs:
        """Test program with short arguments."""

        x: int = 5
        """Short description."""

        y: str = "hello"
        """Another short one."""

    # Test compact mode.
    compact_helptext = get_helptext_with_checks(ShortArgs, compact_help=True)
    assert "--help-verbose" in compact_helptext
    assert "--x INT" in compact_helptext
    assert "--y STR" in compact_helptext
    # In compact mode, primary help text is omitted, only defaults shown.
    assert "Short description" not in compact_helptext
    assert "(default: 5)" in compact_helptext
    assert "(default: hello)" in compact_helptext


def test_compact_help_long_text() -> None:
    """Test compact mode with very long argument names and behavior hints."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class LongArgs:
        """Test program with very long argument names."""

        very_long_argument_name_that_takes_up_a_lot_of_terminal_space: int = 5
        """This is an extremely long description that will definitely exceed the terminal width when combined with the argument name. It contains many words and should trigger the line-splitting behavior in compact mode."""

    # Test compact mode.
    compact_helptext = get_helptext_with_checks(LongArgs, compact_help=True)
    assert (
        "--very-long-argument-name-that-takes-up-a-lot-of-terminal-space"
        in compact_helptext
    )
    # In compact mode, primary descriptions are omitted.
    assert "extremely long description" not in compact_helptext
    # But default values are still shown.
    assert "(default: 5)" in compact_helptext


def test_verbose_help_short_invocation() -> None:
    """Test verbose mode with short invocation names."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class VerboseArgs:
        """Test program for verbose mode."""

        x: int = 5
        """This is a longer description that will appear in verbose mode. It provides detailed information about what the x parameter does and how it should be used."""

        y: str = "test"
        """Another detailed description for the y parameter."""

    # Test verbose mode (default behavior).
    verbose_helptext = get_helptext_with_checks(VerboseArgs, compact_help=False)
    assert "--help-verbose" not in verbose_helptext
    assert "--x INT" in verbose_helptext
    assert "--y STR" in verbose_helptext
    assert "longer description" in verbose_helptext
    assert "Another detailed description" in verbose_helptext


def test_verbose_help_long_invocation() -> None:
    """Test verbose mode with very long invocation names that exceed max width."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class VeryLongInvocationArgs:
        """Test program with very long argument names."""

        extremely_long_argument_name_that_exceeds_maximum_width_threshold: int = 5
        """Description for the long argument."""

        another_incredibly_long_argument_name_for_testing_purposes: str = "test"
        """Another description."""

    # Test verbose mode with long invocations.
    verbose_helptext = get_helptext_with_checks(
        VeryLongInvocationArgs, compact_help=False
    )
    assert (
        "--extremely-long-argument-name-that-exceeds-maximum-width-threshold"
        in verbose_helptext
    )
    assert (
        "--another-incredibly-long-argument-name-for-testing-purposes"
        in verbose_helptext
    )
    assert "Description for the long argument" in verbose_helptext


def test_compact_vs_verbose_help() -> None:
    """Test that compact and verbose modes produce different output."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class CompareArgs:
        """Test comparing compact vs verbose."""

        x: int = 5
        """A description for parameter x."""

    compact = get_helptext_with_checks(CompareArgs, compact_help=True)
    verbose = get_helptext_with_checks(CompareArgs, compact_help=False)

    # Compact should have --help-verbose.
    assert "--help-verbose" in compact
    assert "--help-verbose" not in verbose

    # Both should have the basic argument.
    assert "--x INT" in compact
    assert "--x INT" in verbose


def test_no_duplicate_description() -> None:
    """Test that description passed via description= kwarg only appears once.

    Regression test for issue where description would appear both after usage line
    and inside the options box, causing duplication.
    """

    def main(x: int) -> None:
        pass

    # Test with description passed as keyword argument.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(
            main, args=["--help"], description="This helptext should only appear once."
        )

    helptext = tyro._strings.strip_ansi_sequences(target.getvalue())
    # The description should appear in the helptext.
    assert "This helptext should only appear once." in helptext
    # But it should only appear once, not twice.
    assert helptext.count("This helptext should only appear once.") == 1


def test_literal_invalid_choice_error_message() -> None:
    """Test that invalid choice error for Literal types shows clean argument name.

    Regression test for issue where internal implementation detail
    '__tyro-dummy-inner__' was exposed in error messages.
    """

    # When using a Literal type as the main CLI type, invalid choice errors
    # should show a clean argument name, not internal implementation details.
    # This will also be caught by the assertion in helptext_utils, but we test
    # it explicitly here for error messages.
    # Capture stderr to check the error message.
    stderr = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stderr(stderr):
            tyro.cli(Literal["a", "b"], args=["c"], return_unknown_args=True)  # type: ignore

    # Check the error message.
    error_message = stderr.getvalue()
    # The error message should mention the invalid choice 'c'.
    assert "invalid choice" in error_message.lower() or "c" in error_message
    # But it should NOT expose internal implementation details.
    assert "__tyro-dummy-inner__" not in error_message


def test_optional_union_subparser_helptext() -> None:
    """Test that optional union subparsers are marked as optional in help text."""

    @dataclasses.dataclass
    class StructA:
        """First option."""

        a: int = 1

    @dataclasses.dataclass
    class StructB:
        """Second option."""

        b: int = 2

    # Optional union (total=False).
    class OptionalConfig(TypedDict, total=False):
        """Config with optional union field."""

        choice: Union[StructA, StructB]
        other: int

    helptext_optional = get_helptext_with_checks(OptionalConfig)

    # Check that "(required)" is NOT shown for optional subparser.
    lines = helptext_optional.split("\n")
    in_subcommands_section = False
    for line in lines:
        if "subcommands" in line.lower():
            in_subcommands_section = True
        if in_subcommands_section and "choice:struct-a" in line:
            # The section header before the subcommand list should not say "(required)".
            assert (
                "(required)"
                not in helptext_optional.split("choice:struct-a")[0].split(
                    "subcommands"
                )[-1]
            )
            break

    # For tyro backend, check usage line shows optional subparser in brackets.
    # For argparse backend, the brackets may not show up in the usage line due to
    # argparse limitations, but "optional" should appear in the section title.
    if "[{choice:struct-a,choice:struct-b}]" not in helptext_optional:
        # Argparse backend - check for "optional" in section title.
        assert (
            "optional" in helptext_optional.lower()
            or "choice subcommands" in helptext_optional
        )

    # Required union (total=True by default).
    class RequiredConfig(TypedDict):
        """Config with required union field."""

        choice: Union[StructA, StructB]
        other: int

    helptext_required = get_helptext_with_checks(RequiredConfig)

    # Check usage line shows required subparser.
    assert "{choice:struct-a,choice:struct-b}" in helptext_required

    # Check that "(required)" IS shown for required subparser (tyro backend).
    # Argparse backend doesn't show "(required)" but doesn't mark subcommands as optional.
    if "(required)" not in helptext_required:
        # Argparse backend - ensure "optional subcommands" is NOT in the help text.
        # (The word "optional" may appear elsewhere, like in field descriptions)
        assert "optional subcommands" not in helptext_required.lower()


def test_subcommand_help_with_required_parent_args() -> None:
    """Test that subcommand --help works without satisfying parent required args.

    This is a regression test for https://github.com/brentyi/tyro/issues/403
    where using `python script.py sub:a --help` would fail with "Missing required
    argument: --x" instead of showing the subcommand help.
    """

    @dataclasses.dataclass
    class A:
        a: int

    @dataclasses.dataclass
    class B:
        b: int

    def main(x: int, sub: Union[A, B]) -> None:
        pass

    # This should show help text for the subcommand without requiring --x.
    with pytest.raises(SystemExit) as exc_info:
        tyro.cli(main, args=["sub:a", "--help"])
    assert exc_info.value.code == 0

    # Verify the help text is actually shown (not an error).
    captured = io.StringIO()
    with contextlib.redirect_stdout(captured):
        with pytest.raises(SystemExit):
            tyro.cli(main, args=["sub:a", "--help"])
    helptext = captured.getvalue()
    assert "sub:a" in helptext or "--sub.a" in helptext

    # Also test with -h flag.
    with pytest.raises(SystemExit) as exc_info:
        tyro.cli(main, args=["sub:a", "-h"])
    assert exc_info.value.code == 0

    # -H and --help-verbose are tyro-backend-only features, and only work
    # when compact_help=True.
    if tyro._experimental_options["backend"] == "tyro":
        # Test with compact_help=True: all help flags should work.
        with pytest.raises(SystemExit) as exc_info:
            tyro.cli(main, args=["sub:a", "--help"], compact_help=True)
        assert exc_info.value.code == 0

        with pytest.raises(SystemExit) as exc_info:
            tyro.cli(main, args=["sub:a", "-h"], compact_help=True)
        assert exc_info.value.code == 0

        with pytest.raises(SystemExit) as exc_info:
            tyro.cli(main, args=["sub:a", "-H"], compact_help=True)
        assert exc_info.value.code == 0

        with pytest.raises(SystemExit) as exc_info:
            tyro.cli(main, args=["sub:a", "--help-verbose"], compact_help=True)
        assert exc_info.value.code == 0


def test_comment_not_leaking_through_non_field_lines() -> None:
    """Comments above non-field lines should not leak to subsequent fields."""

    @dataclasses.dataclass
    class Config:
        x: int = 1
        # This comment is for the internal variable.
        _internal = "something"
        y: int = 2

    helptext = get_helptext_with_checks(Config)
    assert "This comment is for the internal variable" not in helptext



from __future__ import annotations

import sys
from typing import Any, Tuple, Type


def is_flax_module(cls: Type[Any]) -> Tuple[bool, Tuple[str, ...]]:
    """Check if a class is a Flax module and return fields to skip.

    Returns:
        Tuple of (is_flax_module, fields_to_skip)
    """
    # Skip non-dataclasses
    if not is_dataclass_type(cls):
        return False, ()

    # Check if dataclass is a flax module. This is only possible if flax is already
    # loaded.
    #
    # We generally want to avoid importing flax, since it requires a lot of heavy
    # imports.
    if "flax.linen" in sys.modules.keys():
        try:
            import flax.linen

            if issubclass(cls, flax.linen.Module):
                # For flax modules, we ignore the built-in "name" and "parent" fields.
                return True, ("name", "parent")
        except ImportError:
            pass

    return False, ()


def is_dataclass_type(obj: Any) -> bool:
    """Check if an object is a dataclass type."""
    import dataclasses

    return isinstance(obj, type) and dataclasses.is_dataclass(obj)



"""Tests for bugs in _calling.py argument handling and error reporting."""

import contextlib
import io

import pytest

import tyro
from tyro._strings import strip_ansi_sequences


def test_optional_group_missing_args_reports_all_missing() -> None:
    """When some args in an optional group are provided and others are missing,
    the error message should list ALL missing arguments, not just those before
    the first non-missing one.

    Bug: the loop in _calling.py used ``break`` instead of ``continue`` when
    encountering a non-missing kwarg, so missing args after it were silently
    dropped from the error message.
    """

    class Coord:
        """A coordinate with three required fields."""

        def __init__(self, a: int, b: int, c: int):
            self.a = a
            self.b = b
            self.c = c

    def main(coord: Coord = Coord(a=1, b=2, c=3)) -> Coord:
        return coord

    # Provide only --coord.b, leaving a and c unprovided.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(main, args=["--coord.b", "20"])

    error = strip_ansi_sequences(target.getvalue())
    # Both --coord.a and --coord.c should be reported as missing.
    assert "--coord.a" in error, f"Expected --coord.a in error message, got: {error}"
    assert "--coord.c" in error, f"Expected --coord.c in error message, got: {error}"


def test_valueerror_no_message_in_constructor() -> None:
    """When a nested struct constructor raises ValueError() with no message,
    tyro should produce a clean error instead of crashing with IndexError.

    Bug: ``e.args[0]`` in _calling.py crashes with IndexError when
    ValueError() is raised with no arguments (empty args tuple).
    """

    class Validated:
        """A type whose constructor raises bare ValueError."""

        def __init__(self, x: int):
            if x > 10:
                raise ValueError()
            self.x = x

    def main(val: Validated = Validated(x=5)) -> Validated:
        return val

    # Should produce a SystemExit with a clean error, not an IndexError.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--val.x", "20"])



"""Tests for unions of Annotated types with subcommand_type_from_defaults constructors.

Tests that Annotated[T, tyro.conf.arg(constructor=subcommand_type_from_defaults(...))]
in a union correctly flattens inner subcommands into the outer union.
"""

import dataclasses
from typing import Optional, Union

from helptext_utils import get_helptext_with_checks
from typing_extensions import Annotated

import tyro


@dataclasses.dataclass(frozen=True)
class Config:
    name: str
    value: int


@dataclasses.dataclass(frozen=True)
class OptimizerA:
    lr: float = 0.01


@dataclasses.dataclass(frozen=True)
class OptimizerB:
    momentum: float = 0.9


@dataclasses.dataclass(frozen=True)
class PlainConfig:
    x: int = 0


@dataclasses.dataclass(frozen=True)
class Inner:
    name: str
    value: int


@dataclasses.dataclass(frozen=True)
class Outer:
    label: str
    inner: Union[
        Annotated[
            Inner,
            tyro.conf.arg(
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {
                        "small": Inner(name="small", value=1),
                        "large": Inner(name="large", value=100),
                    }
                )
            ),
        ],
        Annotated[
            Inner,
            tyro.conf.arg(
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {
                        "custom": Inner(name="custom", value=50),
                    }
                )
            ),
        ],
    ]


def test_basic_annotated_union() -> None:
    """Basic case: union of two annotated configs + None."""

    @dataclasses.dataclass(frozen=True)
    class Args:
        config: Union[
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {
                            "a": Config(name="a", value=1),
                            "b": Config(name="b", value=2),
                        }
                    )
                ),
            ],
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {
                            "c": Config(name="c", value=3),
                            "d": Config(name="d", value=4),
                        }
                    )
                ),
            ],
            None,
        ]

    # Verify all subcommands appear in help.
    helptext = get_helptext_with_checks(Args)
    assert "config:a" in helptext
    assert "config:b" in helptext
    assert "config:c" in helptext
    assert "config:d" in helptext
    assert "config:none" in helptext

    # Verify each subcommand parses correctly.
    assert tyro.cli(Args, args=["config:a"]) == Args(config=Config(name="a", value=1))
    assert tyro.cli(Args, args=["config:b"]) == Args(config=Config(name="b", value=2))
    assert tyro.cli(Args, args=["config:c"]) == Args(config=Config(name="c", value=3))
    assert tyro.cli(Args, args=["config:d"]) == Args(config=Config(name="d", value=4))
    assert tyro.cli(Args, args=["config:none"]) == Args(config=None)


def test_annotated_union_no_none() -> None:
    """Union of two annotated configs without None."""

    @dataclasses.dataclass(frozen=True)
    class Args:
        config: Union[
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {"x": Config(name="x", value=10)}
                    )
                ),
            ],
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {"y": Config(name="y", value=20)}
                    )
                ),
            ],
        ]

    assert tyro.cli(Args, args=["config:x"]) == Args(config=Config(name="x", value=10))
    assert tyro.cli(Args, args=["config:y"]) == Args(config=Config(name="y", value=20))


def test_annotated_union_single() -> None:
    """Single annotated config in a union with None."""

    @dataclasses.dataclass(frozen=True)
    class Args:
        config: Union[
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {
                            "p": Config(name="p", value=100),
                            "q": Config(name="q", value=200),
                        }
                    )
                ),
            ],
            None,
        ]

    helptext = get_helptext_with_checks(Args)
    assert "config:p" in helptext
    assert "config:q" in helptext
    assert "config:none" in helptext

    assert tyro.cli(Args, args=["config:p"]) == Args(config=Config(name="p", value=100))
    assert tyro.cli(Args, args=["config:q"]) == Args(config=Config(name="q", value=200))
    assert tyro.cli(Args, args=["config:none"]) == Args(config=None)


def test_annotated_union_with_override() -> None:
    """Test that subcommand defaults can be overridden via CLI args."""

    @dataclasses.dataclass(frozen=True)
    class Args:
        config: Union[
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {"first": Config(name="first", value=1)}
                    )
                ),
            ],
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {"second": Config(name="second", value=2)}
                    )
                ),
            ],
        ]

    result = tyro.cli(
        Args,
        args=["config:first", "--config.name", "custom", "--config.value", "99"],
    )
    assert result == Args(config=Config(name="custom", value=99))

    result = tyro.cli(
        Args,
        args=["config:second", "--config.name", "other", "--config.value", "42"],
    )
    assert result == Args(config=Config(name="other", value=42))


def test_annotated_union_different_types() -> None:
    """Union of annotated configs with different underlying types."""

    @dataclasses.dataclass(frozen=True)
    class TrainArgs:
        opt: Union[
            Annotated[
                OptimizerA,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {
                            "adam": OptimizerA(lr=0.001),
                            "sgd-fast": OptimizerA(lr=0.1),
                        }
                    )
                ),
            ],
            Annotated[
                OptimizerB,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {"heavy": OptimizerB(momentum=0.99)}
                    )
                ),
            ],
        ]

    assert tyro.cli(TrainArgs, args=["opt:adam"]) == TrainArgs(opt=OptimizerA(lr=0.001))
    assert tyro.cli(TrainArgs, args=["opt:sgd-fast"]) == TrainArgs(
        opt=OptimizerA(lr=0.1)
    )
    assert tyro.cli(TrainArgs, args=["opt:heavy"]) == TrainArgs(
        opt=OptimizerB(momentum=0.99)
    )


def test_annotated_union_three_groups() -> None:
    """Three groups of annotated configs in a union."""

    @dataclasses.dataclass(frozen=True)
    class Args:
        config: Union[
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {
                            "a1": Config(name="a1", value=1),
                            "a2": Config(name="a2", value=2),
                        }
                    )
                ),
            ],
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {"b1": Config(name="b1", value=3)}
                    )
                ),
            ],
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {
                            "c1": Config(name="c1", value=4),
                            "c2": Config(name="c2", value=5),
                            "c3": Config(name="c3", value=6),
                        }
                    )
                ),
            ],
        ]

    helptext = get_helptext_with_checks(Args)
    for name in (
        "config:a1",
        "config:a2",
        "config:b1",
        "config:c1",
        "config:c2",
        "config:c3",
    ):
        assert name in helptext

    assert tyro.cli(Args, args=["config:a1"]) == Args(config=Config(name="a1", value=1))
    assert tyro.cli(Args, args=["config:c3"]) == Args(config=Config(name="c3", value=6))


def test_annotated_union_mixed_with_plain_struct() -> None:
    """Mix of annotated config (with subcommand_type_from_defaults) and a plain struct type."""

    @dataclasses.dataclass(frozen=True)
    class Args:
        config: Union[
            Annotated[
                Config,
                tyro.conf.arg(
                    constructor=tyro.extras.subcommand_type_from_defaults(
                        {
                            "preset-a": Config(name="preset-a", value=1),
                            "preset-b": Config(name="preset-b", value=2),
                        }
                    )
                ),
            ],
            PlainConfig,
        ]

    helptext = get_helptext_with_checks(Args)
    assert "config:preset-a" in helptext
    assert "config:preset-b" in helptext
    assert "config:plain-config" in helptext

    assert tyro.cli(Args, args=["config:preset-a"]) == Args(
        config=Config(name="preset-a", value=1)
    )
    assert tyro.cli(Args, args=["config:plain-config", "--config.x", "42"]) == Args(
        config=PlainConfig(x=42)
    )


def test_annotated_union_nested_struct() -> None:
    """Annotated union inside a nested struct."""

    @dataclasses.dataclass(frozen=True)
    class Args:
        outer: Outer

    result = tyro.cli(Args, args=["--outer.label", "test", "outer.inner:small"])
    assert result == Args(outer=Outer(label="test", inner=Inner(name="small", value=1)))

    result = tyro.cli(Args, args=["--outer.label", "test", "outer.inner:custom"])
    assert result == Args(
        outer=Outer(label="test", inner=Inner(name="custom", value=50))
    )


def test_prefix_name_false_on_subcommands() -> None:
    """prefix_name=False on tyro.conf.arg() should remove the field prefix from subcommand names."""

    @dataclasses.dataclass(frozen=True)
    class Args:
        main: Annotated[
            Config,
            tyro.conf.arg(
                prefix_name=False,
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {
                        "a": Config(name="a", value=1),
                        "b": Config(name="b", value=2),
                    }
                ),
            ),
        ]

    helptext = get_helptext_with_checks(Args)
    # Should be "a" and "b", not "main:a" and "main:b".
    assert "a" in helptext
    assert "b" in helptext
    assert "main:a" not in helptext
    assert "main:b" not in helptext

    assert tyro.cli(Args, args=["a"]) == Args(main=Config(name="a", value=1))
    assert tyro.cli(Args, args=["b"]) == Args(main=Config(name="b", value=2))


def test_prefix_name_false_with_annotated_union() -> None:
    """prefix_name=False combined with annotated union flattening."""

    @dataclasses.dataclass(frozen=True)
    class Args:
        config: Annotated[
            Union[
                Annotated[
                    Config,
                    tyro.conf.arg(
                        constructor=tyro.extras.subcommand_type_from_defaults(
                            {"x": Config(name="x", value=10)}
                        )
                    ),
                ],
                Annotated[
                    Config,
                    tyro.conf.arg(
                        constructor=tyro.extras.subcommand_type_from_defaults(
                            {"y": Config(name="y", value=20)}
                        )
                    ),
                ],
                None,
            ],
            tyro.conf.arg(prefix_name=False),
        ]

    helptext = get_helptext_with_checks(Args)
    assert "x" in helptext
    assert "y" in helptext
    assert "none" in helptext
    assert "config:x" not in helptext
    assert "config:y" not in helptext

    assert tyro.cli(Args, args=["x"]) == Args(config=Config(name="x", value=10))
    assert tyro.cli(Args, args=["y"]) == Args(config=Config(name="y", value=20))
    assert tyro.cli(Args, args=["none"]) == Args(config=None)


def test_prefix_name_false_alongside_prefixed() -> None:
    """One field with prefix_name=False, another with default prefix_name=True."""

    AnnotatedInferenceConfig = Annotated[
        Config,
        tyro.conf.arg(
            constructor=tyro.extras.subcommand_type_from_defaults(
                {
                    "a": Config(name="a", value=1),
                    "b": Config(name="b", value=2),
                }
            )
        ),
    ]

    Args = dataclasses.make_dataclass(
        "Args",
        [
            (
                "main",
                Annotated[AnnotatedInferenceConfig, tyro.conf.arg(prefix_name=False)],
            ),
            ("secondary", Optional[AnnotatedInferenceConfig]),
        ],
        frozen=True,
    )

    # Top-level help: main subcommands should be unprefixed.
    helptext = get_helptext_with_checks(Args)
    # main subcommands should NOT have "main:" prefix.
    assert "main:a" not in helptext
    assert "main:b" not in helptext

    result = tyro.cli(Args, args=["a", "secondary:b"])  # type: ignore[var-annotated]
    assert result == Args(
        main=Config(name="a", value=1), secondary=Config(name="b", value=2)
    )

    result = tyro.cli(Args, args=["b", "secondary:none"])
    assert result == Args(main=Config(name="b", value=2), secondary=None)


def test_prefix_name_false_plain_union() -> None:
    """prefix_name=False on a plain union (no subcommand_type_from_defaults)."""

    @dataclasses.dataclass(frozen=True)
    class Foo:
        x: int = 1

    @dataclasses.dataclass(frozen=True)
    class Bar:
        y: str = "hi"

    @dataclasses.dataclass(frozen=True)
    class Args:
        thing: Annotated[Union[Foo, Bar], tyro.conf.arg(prefix_name=False)]

    helptext = get_helptext_with_checks(Args)
    assert "foo" in helptext
    assert "bar" in helptext
    assert "thing:foo" not in helptext
    assert "thing:bar" not in helptext

    assert tyro.cli(Args, args=["foo"]) == Args(thing=Foo(x=1))
    assert tyro.cli(Args, args=["bar"]) == Args(thing=Bar(y="hi"))



# mypy: ignore-errors
#
# Passing a Union type directly to tyro.cli() doesn't type-check correctly in
# mypy. This will be fixed by `typing.TypeForm`: https://peps.python.org/pep-0747/
"""Generic subcommands

Just like standard classes, generic classes within unions can be selected
between using subcommands.

Usage:

    python ./03_generic_subcommands.py --help
    python ./03_generic_subcommands.py experiment-adam --help
    python ./03_generic_subcommands.py experiment-sgd --help
    python ./03_generic_subcommands.py experiment-adam --path /tmp --opt.lr 1e-3
"""

import dataclasses
from pathlib import Path

import tyro


@dataclasses.dataclass
class Sgd:
    lr: float = 1e-4


@dataclasses.dataclass
class Adam:
    lr: float = 3e-4
    betas: tuple[float, float] = (0.9, 0.999)


@dataclasses.dataclass
class Experiment[OptimizerT: (Adam, Sgd)]:
    path: Path
    opt: OptimizerT


if __name__ == "__main__":
    args = tyro.cli(Experiment[Adam] | Experiment[Sgd])
    print(args)



"""Tests for subcommand default matching with numeric type coercion.

Regression tests for:
1. int values should match float type annotations (numeric tower).
2. none-type should not appear in error messages from subcommand_type_from_defaults.
"""

from __future__ import annotations

import dataclasses
from typing import Tuple

import pytest

import tyro


@dataclasses.dataclass
class Optimizer:
    lr: float = 1.0


@dataclasses.dataclass
class Config:
    optimizers: dict[str, Optimizer]


def test_subcommand_int_matching_float() -> None:
    """int default values should match float type annotations."""
    result = tyro.cli(
        tyro.extras.subcommand_type_from_defaults(
            {
                "a": Config({"embeddings": Optimizer(lr=1)}),
                "b": Config({"embeddings": Optimizer(lr=2)}),
            }
        ),
        default=Config({"embeddings": Optimizer(lr=3)}),
        args=[],
    )
    assert isinstance(result, Config)


def test_subcommand_int_matching_float_select_a() -> None:
    """Selecting subcommand 'a' should work with int defaults for float fields."""
    result = tyro.cli(
        tyro.extras.subcommand_type_from_defaults(
            {
                "a": Config({"embeddings": Optimizer(lr=1)}),
                "b": Config({"embeddings": Optimizer(lr=2)}),
            }
        ),
        default=Config({"embeddings": Optimizer(lr=3)}),
        args=["a"],
    )
    assert isinstance(result, Config)
    # The overall default (lr=3) overrides the subcommand's default (lr=1).
    assert result.optimizers["embeddings"].lr == 3


@dataclasses.dataclass
class TupleConfig:
    point: Tuple[float, float] = (0.0, 0.0)


def test_subcommand_tuple_float_with_int_defaults() -> None:
    """tuple[float, float] fields with int default values should match."""
    result = tyro.cli(
        tyro.extras.subcommand_type_from_defaults(
            {
                "a": TupleConfig(point=(1, 2)),
                "b": TupleConfig(point=(2, 3)),
                "c": TupleConfig(point=(3, 4)),
            }
        ),
        default=TupleConfig(point=(1, 2)),
        args=[],
    )
    assert isinstance(result, TupleConfig)


def test_subcommand_tuple_float_with_int_defaults_select() -> None:
    """Selecting a specific subcommand with int defaults for float tuple fields."""
    result = tyro.cli(
        tyro.extras.subcommand_type_from_defaults(
            {
                "a": TupleConfig(point=(1, 2)),
                "b": TupleConfig(point=(2, 3)),
                "c": TupleConfig(point=(3, 4)),
            }
        ),
        default=TupleConfig(point=(1, 2)),
        args=["b"],
    )
    assert isinstance(result, TupleConfig)
    assert result.point == (2.0, 3.0)


def test_subcommand_tuple_float_with_int_defaults_override() -> None:
    """Overriding tuple[float, float] values via CLI args."""
    result = tyro.cli(
        tyro.extras.subcommand_type_from_defaults(
            {
                "a": TupleConfig(point=(1, 2)),
                "b": TupleConfig(point=(2, 3)),
                "c": TupleConfig(point=(3, 4)),
            }
        ),
        default=TupleConfig(point=(1, 2)),
        args=["c", "--point", "5.0", "6.0"],
    )
    assert isinstance(result, TupleConfig)
    assert result.point == (5.0, 6.0)


@dataclasses.dataclass
class NestedTupleConfig:
    params: dict[str, TupleConfig]


def test_subcommand_nested_tuple_float_with_int_defaults() -> None:
    """Nested dict with tuple[float, float] fields and int defaults should match."""
    result = tyro.cli(
        tyro.extras.subcommand_type_from_defaults(
            {
                "a": NestedTupleConfig(params={"layer1": TupleConfig(point=(1, 2))}),
                "b": NestedTupleConfig(params={"layer1": TupleConfig(point=(2, 3))}),
            }
        ),
        default=NestedTupleConfig(params={"layer1": TupleConfig(point=(3, 4))}),
        args=[],
    )
    assert isinstance(result, NestedTupleConfig)


@dataclasses.dataclass
class StrictConfig:
    """Config where passing wrong-typed defaults causes structural mismatch."""

    name: str = "hello"


def test_subcommand_unmatched_default_no_none_type_in_error(capsys) -> None:
    """When no subcommand matches the default, the error should not mention none-type.

    Regression test: subcommand_type_from_defaults pads with Annotated[None, Suppress]
    for single-element unions. This padding should not appear in error messages.
    """
    # Create a default where the field value has the wrong type (int for str field).
    # The type of the instance matches (StrictConfig), but the field value doesn't
    # pass structural matching.
    bad_default = StrictConfig(name=42)  # type: ignore
    with pytest.raises(SystemExit):
        tyro.cli(
            tyro.extras.subcommand_type_from_defaults(
                {
                    "good": StrictConfig(name="world"),
                }
            ),
            default=bad_default,
            args=[],
        )
    captured = capsys.readouterr()
    assert "none-type" not in captured.err, (
        "Suppressed NoneType padding should not appear in error messages"
    )
    assert "good" in captured.err, "Real subcommand should appear in error"


def test_subcommand_multi_unmatched_default_no_none_type_in_error(capsys) -> None:
    """When no subcommand matches, none-type should not appear even with multiple subcommands."""
    bad_default = StrictConfig(name=42)  # type: ignore
    with pytest.raises(SystemExit):
        tyro.cli(
            tyro.extras.subcommand_type_from_defaults(
                {
                    "x": StrictConfig(name="a"),
                    "y": StrictConfig(name="b"),
                }
            ),
            default=bad_default,
            args=[],
        )
    captured = capsys.readouterr()
    assert "none-type" not in captured.err
    # Real subcommands should still appear in the error.
    assert "x" in captured.err
    assert "y" in captured.err



"""Booleans and flags

Booleans can either be expected to be explicitly passed in, or, if given a default
value, automatically converted to flags.

To turn off conversion, see :class:`tyro.conf.FlagConversionOff`.

Usage:

    python ./04_flags.py --help
    python ./04_flags.py --boolean True
    python ./04_flags.py --boolean False --flag-a
    python ./04_flags.py --boolean False --no-flag-b
"""

from dataclasses import dataclass
from pprint import pprint

import tyro


@dataclass
class Args:
    # Boolean. This expects an explicit "True" or "False".
    boolean: bool

    # Optional boolean. Same as above, but can be omitted.
    optional_boolean: bool | None = None

    # Pass --flag-a in to set this value to True.
    flag_a: bool = False

    # Pass --no-flag-b in to set this value to False.
    flag_b: bool = True


if __name__ == "__main__":
    args = tyro.cli(Args)
    pprint(args)



"""Custom primitive

In this example, we use :mod:`tyro.constructors` to attach a primitive
constructor via a runtime annotation.

Usage:

    python ./02_primitive_annotation.py --help
    python ./02_primitive_annotation.py --dict1 '{"hello": "world"}'
    python ./02_primitive_annotation.py --dict1 '{"hello": "world"}' --dict2 '{"hello": "world"}'
"""

import json

from typing_extensions import Annotated

import tyro

# A dictionary type, but `tyro` will expect a JSON string from the CLI.
JsonDict = Annotated[
    dict,
    tyro.constructors.PrimitiveConstructorSpec(
        # Number of arguments to consume.
        nargs=1,
        # Argument name in usage messages.
        metavar="JSON",
        # Convert a list of strings to an instance. The length of the list
        # should match `nargs`.
        instance_from_str=lambda args: json.loads(args[0]),
        # Check if an instance is of the expected type. This is only used for
        # helptext formatting in the presence of union types.
        is_instance=lambda instance: isinstance(instance, dict),
        # Convert an instance to a list of strings. This is used for handling
        # default values that are set in Python. The length of the list should
        # match `nargs`.
        str_from_instance=lambda instance: [json.dumps(instance)],
    ),
]


def main(
    dict1: JsonDict,
    dict2: JsonDict = {"default": None},
) -> None:
    print(f"{dict1=}")
    print(f"{dict2=}")


if __name__ == "__main__":
    tyro.cli(main)



"""The :mod:`tyro.extras` submodule contains helpers that complement :func:`tyro.cli()`.

.. warning::

    Compared to the core interface, APIs here are more likely to be changed or deprecated.

"""

from .._cli import get_parser as get_parser
from .._settings import set_accent_color as set_accent_color
from ._base_configs import overridable_config_cli as overridable_config_cli
from ._base_configs import (
    subcommand_type_from_defaults as subcommand_type_from_defaults,
)
from ._choices_type import literal_type_from_choices as literal_type_from_choices
from ._serialization import from_yaml as from_yaml
from ._serialization import to_yaml as to_yaml
from ._subcommand_app import SubcommandApp as SubcommandApp
from ._subcommand_cli_from_dict import (
    subcommand_cli_from_dict as subcommand_cli_from_dict,
)



"""PyTorch parallelism

The :code:`console_outputs=` argument can be set to :code:`False` to suppress helptext and
error message printing.

This is useful in PyTorch for distributed training scripts, where you only want
to print helptext from the main process:


.. code-block:: python

    # HuggingFace Accelerate.
    args = tyro.cli(Args, console_outputs=accelerator.is_main_process)

    # PyTorch DDP.
    args = tyro.cli(Args, console_outputs=(rank == 0))

    # PyTorch Lightning.
    args = tyro.cli(Args, console_outputs=trainer.is_global_zero)


Usage:

    python ./01_pytorch_parallelism.py --help
"""

import tyro


def train(foo: int, bar: str) -> None:
    """Description. This should show up in the helptext!"""


if __name__ == "__main__":
    args = tyro.cli(train, console_outputs=False)
    print(args)



"""Backend abstraction for parsing command-line arguments.

The backend interface separates parsing from instantiation. Both backends parse
command-line arguments into string values, which are then passed to _calling.py
for object instantiation. This separation provides several benefits:

1. Clean separation of concerns between parsing and instantiation.
2. Reuse of existing instantiation logic in _calling.py.
3. Ability to handle --help without instantiating objects (avoiding side effects).
4. Easier debugging with inspectable intermediate string values.
5. Consistent interface between different backend implementations.
"""

from __future__ import annotations

import abc
from typing import Any, Literal, Sequence

from tyro._backends._argparse_formatter import TyroArgumentParser

from .. import _parsers


class ParserBackend(abc.ABC):
    """Abstract base class for parser backends.

    All backends follow a two-phase approach:
    1. Parse command-line arguments into string values (handled by the backend).
    2. Instantiate objects from parsed values (handled by _calling.py).

    This design ensures backends are interchangeable and can leverage existing
    instantiation logic without duplication.
    """

    @abc.abstractmethod
    def parse_args(
        self,
        parser_spec: _parsers.ParserSpecification,
        args: Sequence[str],
        prog: str,
        return_unknown_args: bool,
        console_outputs: bool,
        add_help: bool,
        compact_help: bool = False,
    ) -> tuple[dict[str | None, Any], list[str] | None]:
        """Parse command-line arguments using the parser specification.

        Args:
            parser_spec: Specification for the parser structure.
            args: Command-line arguments to parse.
            prog: Program name for help text.
            return_unknown_args: If True, return unknown arguments.
            console_outputs: If True, allow console outputs (help, errors).
            add_help: Whether to enable -h/--help.
            compact_help: Whether to use compact help format (TyroBackend only).

        Returns:
            A tuple of (parsed_values, unknown_args).
            parsed_values is a dict mapping field names to string values or
            lists of strings (for multi-value arguments).
            unknown_args is None unless return_unknown_args is True.
        """
        ...

    @abc.abstractmethod
    def get_parser_for_completion(
        self,
        parser_spec: _parsers.ParserSpecification,
        prog: str | None,
        add_help: bool,
    ) -> TyroArgumentParser:
        """Get a parser object for shell completion generation.

        This is needed for compatibility with shtab completion generation.

        Args:
            parser_spec: Specification for the parser structure.
            prog: Program name for help text.
            add_help: Whether to enable -h/--help.

        Returns:
            A parser object compatible with shtab (typically argparse.ArgumentParser).
        """
        ...

    def generate_completion(
        self,
        parser_spec: _parsers.ParserSpecification,
        prog: str,
        shell: Literal["bash", "zsh", "tcsh"],
        root_prefix: str,
    ) -> str:
        """Generate shell completion script directly from parser specification.

        This method can be overridden by backends to provide native completion
        generation. The default implementation falls back to shtab-based completion
        by building an argparse parser.

        Args:
            parser_spec: Specification for the parser structure.
            prog: Program name.
            shell: Shell type ('bash', 'zsh', or 'tcsh').
            root_prefix: Prefix for completion function names.

        Returns:
            Shell completion script as a string.
        """
        # Default implementation: use shtab with argparse parser.
        try:
            import shtab
        except ImportError as e:  # pragma: no cover
            raise ImportError(
                "shtab is required for completion generation with the argparse backend. "
                "Install it with: pip install shtab>=1.5.6"
            ) from e

        parser = self.get_parser_for_completion(parser_spec, prog=prog, add_help=True)
        return shtab.complete(parser=parser, shell=shell, root_prefix=root_prefix)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

import pytest
from helptext_utils import get_helptext_with_checks
from typing_extensions import Annotated

import tyro


def test_positional():
    def main(
        x: int,
        y: int,
        /,
        # Note: it's generally a bad idea to have a mutable object (like a list) as a
        # default value. But it should still work.
        z: List[int] = [1, 2, 3],
    ) -> Tuple[int, int, int]:
        """main.

        Args:
            x: x
            y: y
            z: z

        Returns:
            Tuple[int, int, int]: Output.
        """
        return (x, y, z[0])

    assert tyro.cli(main, args="1 2 --z 3".split(" ")) == (1, 2, 3)
    with pytest.raises(SystemExit):
        assert tyro.cli(main, args="--x 1 --y 2 --z 3".split(" ")) == (1, 2, 3)


class A:
    def __init__(self, a: int, hello_world: int, /, c: int):
        self.hello_world = hello_world


def test_nested_positional():
    def nest1(a: int, b: int, thing: A, /, c: int) -> A:
        return thing

    assert isinstance(tyro.cli(nest1, args="0 1 2 3 4 --c 4".split(" ")), A)
    assert tyro.cli(nest1, args="0 1 2 3 4 --c 4".split(" ")).hello_world == 3
    with pytest.raises(SystemExit):
        tyro.cli(nest1, args="0 1 2 3 4 4 --c 4".split(" "))


class B:
    def __init__(self, a: int, b: int, /, c: int):
        pass


def test_nested_positional_alt():
    def nest2(a: int, b: int, /, thing: B, c: int):
        return thing

    assert isinstance(tyro.cli(nest2, args="0 1 2 3 --thing.c 4 --c 4".split(" ")), B)
    with pytest.raises(SystemExit):
        tyro.cli(nest2, args="0 1 2 3 4 --thing.c 4 --c 4".split(" "))


def test_positional_with_underscores():
    """Hyphen replacement works a bit different for positional arguments."""

    def main(a_multi_word_input: int, /) -> int:
        return a_multi_word_input

    assert tyro.cli(main, args=["5"]) == 5


def test_positional_booleans():
    """Make sure that flag behavior is disabled for positional booleans."""

    def main(
        flag1: bool,
        flag2: bool = True,
        flag3: bool = False,
        /,
    ) -> Tuple[bool, bool, bool]:
        return flag1, flag2, flag3

    assert tyro.cli(main, args=["True"]) == (True, True, False)
    assert tyro.cli(main, args=["True", "False"]) == (True, False, False)
    assert tyro.cli(main, args=["False", "False", "True"]) == (False, False, True)

    with pytest.raises(SystemExit):
        tyro.cli(main, args=["hmm"])
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["true"])
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["True", "false"])


def test_optional_list():
    def main(a: Optional[List[int]], /) -> Optional[List[int]]:
        return a

    assert tyro.cli(main, args=["None"]) is None
    assert tyro.cli(main, args=["1", "2"]) == [1, 2]
    assert tyro.cli(main, args=[]) == []
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["hm"])


def test_optional_list_with_default():
    def main(a: Optional[List[int]] = None, /) -> Optional[List[int]]:
        return a

    assert tyro.cli(main, args=["None"]) is None
    assert tyro.cli(main, args=["5", "5"]) == [5, 5]
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["None", "5"])


def test_positional_tuple():
    def main(x: Tuple[int, int], y: Tuple[str, str], /):
        return x, y

    assert tyro.cli(main, args="1 2 3 4".split(" ")) == ((1, 2), ("3", "4"))


def make_list_of_strings_with_minimum_length(args: List[str]) -> List[str]:
    if len(args) == 0:
        raise ValueError("Expected at least one string")
    return args


ListOfStringsWithMinimumLength = Annotated[
    List[str],
    tyro.constructors.PrimitiveConstructorSpec(
        nargs="*",
        metavar="STR [STR ...]",
        is_instance=lambda x: (
            isinstance(x, list) and all(isinstance(i, str) for i in x)
        ),
        instance_from_str=make_list_of_strings_with_minimum_length,
        str_from_instance=lambda args: args,
    ),
]


def test_min_length_custom_constructor_positional() -> None:
    def main(
        field1: ListOfStringsWithMinimumLength, /, field2: int = 3
    ) -> ListOfStringsWithMinimumLength:
        del field2
        return field1

    with pytest.raises(SystemExit):
        tyro.cli(main, args=[])
    assert tyro.cli(main, args=["a", "b"]) == ["a", "b"]


TupleCustomConstructor = Annotated[
    Tuple[str, ...],
    tyro.constructors.PrimitiveConstructorSpec(
        nargs="*",
        metavar="A TUPLE METAVAR",
        is_instance=lambda x: (
            isinstance(x, tuple) and all(isinstance(i, str) for i in x)
        ),
        instance_from_str=tuple,
        str_from_instance=list,
    ),
]


def test_tuple_custom_constructors_positional() -> None:
    def main(field1: TupleCustomConstructor, /, field2: int = 3) -> Tuple[str, ...]:
        del field2
        return field1

    assert tyro.cli(main, args=["a", "b"]) == ("a", "b")
    assert tyro.cli(main, args=["a"]) == ("a",)
    assert tyro.cli(main, args=[]) == ()
    assert "A TUPLE METAVAR" in get_helptext_with_checks(main)


TupleCustomConstructor2 = Annotated[
    Tuple[str, ...],
    tyro.constructors.PrimitiveConstructorSpec(
        nargs="*",
        metavar="A TUPLE METAVAR",
        is_instance=lambda x: (
            isinstance(x, tuple) and all(isinstance(i, str) for i in x)
        ),
        instance_from_str=tuple,
        str_from_instance=list,
    ),
]


def test_tuple_custom_constructors_positional_default_none() -> None:
    def main(
        field1: TupleCustomConstructor2 | None = None, /, field2: int = 3
    ) -> Tuple[str, ...] | None:
        del field2
        return field1

    assert tyro.cli(main, args=["a", "b"]) == ("a", "b")
    assert tyro.cli(main, args=["a"]) == ("a",)
    assert tyro.cli(main, args=[]) is None
    assert "A TUPLE METAVAR" in get_helptext_with_checks(main)


def test_tuple_custom_constructors_positional_default_five() -> None:
    def main(
        field1: TupleCustomConstructor2 | int = 5, /, field2: int = 3
    ) -> Tuple[str, ...] | int:
        del field2
        return field1

    assert tyro.cli(main, args=["a", "b"]) == ("a", "b")
    assert tyro.cli(main, args=["a"]) == ("a",)
    assert tyro.cli(main, args=[]) == 5
    assert "A TUPLE METAVAR" in get_helptext_with_checks(main)


@dataclass
class DoubleDashOptions:
    """Test dataclass for double-dash parsing."""

    command: Annotated[
        tyro.conf.Positional[Sequence[str]], tyro.conf.arg(metavar="COMMAND")
    ]
    flag: str = "<unset>"


@dataclass
class DoubleDashFixedPositional:
    """Test dataclass with a fixed (non-sequence) positional argument."""

    command: Annotated[tyro.conf.Positional[str], tyro.conf.arg(metavar="COMMAND")]
    flag: str = "<unset>"


def test_double_dash_end_of_options() -> None:
    """Test that '--' works as end-of-options marker.

    After '--', all arguments should be treated as positional, even if they
    look like flags (e.g., --flag).

    Regression test for: https://github.com/brentyi/tyro/issues/411
    """
    # Basic case: '--' before flag-like positional args.
    result = tyro.cli(DoubleDashOptions, args="test -- --flag flag-test".split())
    assert result.command == ["test", "--flag", "flag-test"]
    assert result.flag == "<unset>"

    # '--' at the beginning: everything is positional.
    result = tyro.cli(DoubleDashOptions, args="-- --flag test".split())
    assert result.command == ["--flag", "test"]
    assert result.flag == "<unset>"

    # '--' at the end (no args after).
    result = tyro.cli(DoubleDashOptions, args="test --".split())
    assert result.command == ["test"]
    assert result.flag == "<unset>"

    # '--' with actual flag before it.
    result = tyro.cli(
        DoubleDashOptions, args="--flag actual-flag test -- --not-a-flag".split()
    )
    assert result.command == ["test", "--not-a-flag"]
    assert result.flag == "actual-flag"

    # '--' after fixed positional with extra args should error (unknown args).
    with pytest.raises(SystemExit):
        tyro.cli(DoubleDashFixedPositional, args="test -- --extra-arg".split())



from __future__ import annotations

import functools
import sys
from typing import TYPE_CHECKING, Any, Dict

from typing_extensions import cast

from .. import _docstrings, _resolver
from .._singleton import MISSING_NONPROP, is_missing
from ._struct_spec import StructConstructorSpec, StructFieldSpec, StructTypeInfo

if TYPE_CHECKING:
    import pydantic as pydantic
    import pydantic.v1.fields as pydantic_v1_fields


def _get_pydantic_v1_field_default(
    name: str,
    field: pydantic_v1_fields.ModelField,
    parent_default_instance: Any,
) -> Any:
    """Helper for getting the default instance for a Pydantic field."""

    # Try grabbing default from parent instance.
    if not is_missing(parent_default_instance) and parent_default_instance is not None:
        # Populate default from some parent, eg `default=` in `tyro.cli()`.
        if hasattr(parent_default_instance, name):
            return getattr(parent_default_instance, name)

    if not field.required:
        return field.get_default()

    # Otherwise, no default.
    return MISSING_NONPROP


def _get_pydantic_v2_field_default(
    name: str,
    field: pydantic.fields.FieldInfo,
    parent_default_instance: Any,
) -> Any:
    """Helper for getting the default instance for a Pydantic field."""

    # Try grabbing default from parent instance.
    if not is_missing(parent_default_instance) and parent_default_instance is not None:
        # Populate default from some parent, eg `default=` in `tyro.cli()`.
        if hasattr(parent_default_instance, name):
            return getattr(parent_default_instance, name)

    if not field.is_required():
        return field.get_default(call_default_factory=True)

    # Otherwise, no default.
    return MISSING_NONPROP


def pydantic_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
    """Rule for handling Pydantic models."""
    # Check if pydantic is imported
    if "pydantic" not in sys.modules.keys():  # pragma: no cover
        return None

    try:
        import pydantic
    except ImportError:
        # Needed for the mock import test in
        # test_missing_optional_packages.py to pass.
        return None

    try:
        if "pydantic.v1" in sys.modules.keys():
            from pydantic import v1 as pydantic_v1
        else:  # pragma: no cover
            pydantic_v1 = None  # type: ignore
    except ImportError:
        pydantic_v1 = None  # type: ignore

    # Check if the type is a Pydantic model
    try:
        if not (
            issubclass(info.type, pydantic.BaseModel)
            or (
                pydantic_v1 is not None and issubclass(info.type, pydantic_v1.BaseModel)
            )
        ):
            return None
    except TypeError:
        # issubclass failed!
        return None

    field_list = []
    pydantic_version = int(getattr(pydantic, "__version__", "1.0.0").partition(".")[0])

    if pydantic_version < 2 or (
        pydantic_v1 is not None and issubclass(info.type, pydantic_v1.BaseModel)
    ):
        # Pydantic 1.xx
        cls_cast = info.type
        hints = _resolver.get_type_hints_resolve_type_params(
            info.type, include_extras=True
        )
        for pd1_field in cast(Dict[str, Any], cls_cast.__fields__).values():
            helptext = pd1_field.field_info.description
            if helptext is None:
                helptext = functools.partial(
                    _docstrings.get_field_docstring,
                    info.type,
                    pd1_field.name,
                    info.markers,
                )

            default = _get_pydantic_v1_field_default(
                pd1_field.name, pd1_field, info.default
            )
            field_list.append(
                StructFieldSpec(
                    name=pd1_field.name,
                    type=hints[pd1_field.name],
                    default=default,
                    helptext=helptext,
                )
            )
    else:
        # Pydantic 2.xx
        hints = _resolver.get_type_hints_resolve_type_params(
            info.type, include_extras=True
        )
        for name, pd2_field in cast(Any, info.type).model_fields.items():
            helptext = pd2_field.description
            if helptext is None:
                helptext = functools.partial(
                    _docstrings.get_field_docstring, info.type, name, info.markers
                )

            default = _get_pydantic_v2_field_default(name, pd2_field, info.default)
            field_list.append(
                StructFieldSpec(
                    name=name,
                    type=hints[name],
                    default=default,
                    helptext=helptext,
                )
            )

    return StructConstructorSpec(instantiate=info.type, fields=tuple(field_list))



"""Tests for nested subparsers functionality."""

from __future__ import annotations

import dataclasses
from typing import Annotated, Any, Union

import pytest

import tyro


@dataclasses.dataclass
class CommandA:
    x: int
    y: str


@dataclasses.dataclass
class CommandB:
    a: float
    b: bool


@dataclasses.dataclass
class CommandC:
    p: list[int]
    q: dict[str, int]


@dataclasses.dataclass
class CommandD:
    value: str


@dataclasses.dataclass
class CommandE:
    count: int


def test_unnamed_nested_union_flattens() -> None:
    """Unnamed nested unions should flatten into a single level."""
    typ: Any = CommandA | Annotated[CommandB | CommandC, None]

    # All three commands should be available at the top level.
    assert tyro.cli(typ, args=["command-a", "--x", "1", "--y", "hello"]) == CommandA(
        1, "hello"
    )
    assert tyro.cli(
        typ,
        args=["command-b", "--a", "2.5", "--b", "True"],
    ) == CommandB(2.5, True)
    assert tyro.cli(
        typ,
        args=[
            "command-c",
            "--p",
            "1",
            "2",
            "--q",
            "k",
            "3",
        ],
    ) == CommandC([1, 2], {"k": 3})


def test_named_nested_union_creates_hierarchy() -> None:
    """Named nested unions should create a hierarchical subparser structure."""
    # A| B| C with name should create nested structure.
    typ: Any = (
        CommandA | Annotated[CommandB | CommandC, tyro.conf.subcommand(name="group-bc")]
    )

    # CommandA should be at top level.
    assert tyro.cli(typ, args=["command-a", "--x", "1", "--y", "hello"]) == CommandA(
        1, "hello"
    )

    # CommandB and CommandC should be under group-bc.
    assert tyro.cli(
        typ, args=["group-bc", "command-b", "--a", "2.5", "--b", "True"]
    ) == CommandB(2.5, True)
    assert tyro.cli(
        typ, args=["group-bc", "command-c", "--p", "1", "2", "--q", "k", "3"]
    ) == CommandC([1, 2], {"k": 3})


def test_deeply_nested_unions() -> None:
    """Test 3+ levels of nesting with mixed named/unnamed unions."""
    # Create a 3-level hierarchy:
    # Top level: command-a or group-rest
    # group-rest: command-b or group-cde
    # group-cde: command-c, command-d, or command-e
    typ: Any = (
        CommandA
        | Annotated[
            CommandB
            | Annotated[
                CommandC | CommandD | CommandE,
                tyro.conf.subcommand(name="group-cde"),
            ],
            tyro.conf.subcommand(name="group-rest"),
        ]
    )

    # Test each path through the hierarchy.
    assert tyro.cli(typ, args=["command-a", "--x", "1", "--y", "hello"]) == CommandA(
        1, "hello"
    )
    assert tyro.cli(
        typ, args=["group-rest", "command-b", "--a", "2.5", "--b", "True"]
    ) == CommandB(2.5, True)
    assert tyro.cli(
        typ, args=["group-rest", "group-cde", "command-c", "--p", "1", "--q", "k", "3"]
    ) == CommandC([1], {"k": 3})
    assert tyro.cli(
        typ, args=["group-rest", "group-cde", "command-d", "--value", "test"]
    ) == CommandD("test")
    assert tyro.cli(
        typ, args=["group-rest", "group-cde", "command-e", "--count", "42"]
    ) == CommandE(42)


def test_mixed_named_unnamed_nested_unions() -> None:
    """Test mixing named and unnamed nested unions - unnamed should flatten into parent."""
    # Create mixed structure:
    # Top: command-a or group-bc or command-d (command-d flattened from unnamed union)
    typ: Any = (
        CommandA
        | Annotated[CommandB | CommandC, tyro.conf.subcommand(name="group-bc")]
        | Annotated[Union[CommandD], None]
    )

    # command-a and command-d at top level, command-b and command-c under group-bc.
    assert tyro.cli(typ, args=["command-a", "--x", "1", "--y", "hello"]) == CommandA(
        1, "hello"
    )
    assert tyro.cli(typ, args=["command-d", "--value", "test"]) == CommandD("test")
    assert tyro.cli(
        typ, args=["group-bc", "command-b", "--a", "2.5", "--b", "True"]
    ) == CommandB(2.5, True)
    assert tyro.cli(
        typ, args=["group-bc", "command-c", "--p", "1", "--q", "k", "3"]
    ) == CommandC([1], {"k": 3})


def test_multiple_named_groups_at_same_level() -> None:
    """Test multiple named subparser groups at the same level."""
    # Workaround for: https://github.com/microsoft/pyright/issues/11046
    # Getting this to pass type checking after the test gen runs and converts
    # `X| Y` types to `X | Y` was trickier than expected...
    AorB: Any = Annotated[CommandA | CommandB, tyro.conf.subcommand(name="group-ab")]
    CorD: Any = Annotated[CommandC | CommandD, tyro.conf.subcommand(name="group-cd")]
    typ: Any = AorB | CorD | CommandC

    assert tyro.cli(
        typ, args=["group-ab", "command-a", "--x", "1", "--y", "hello"]
    ) == CommandA(1, "hello")
    assert tyro.cli(
        typ, args=["group-ab", "command-b", "--a", "2.5", "--b", "True"]
    ) == CommandB(2.5, True)
    assert tyro.cli(
        typ, args=["group-cd", "command-c", "--p", "1", "--q", "k", "3"]
    ) == CommandC([1], {"k": 3})
    assert tyro.cli(typ, args=["command-c", "--p", "1", "--q", "k", "3"]) == CommandC(
        [1], {"k": 3}
    )
    assert tyro.cli(typ, args=["group-cd", "command-d", "--value", "test"]) == CommandD(
        "test"
    )


def test_only_named_groups_at_same_level() -> None:
    """Test multiple named subparser groups at the same level."""
    # Workaround for: https://github.com/microsoft/pyright/issues/11046
    # Getting this to pass type checking after the test gen runs and converts
    # `X| Y` types to `X | Y` was trickier than expected...
    AorB: Any = Annotated[CommandA | CommandB, tyro.conf.subcommand(name="group-ab")]
    CorD: Any = Annotated[CommandC | CommandD, tyro.conf.subcommand(name="group-cd")]
    typ: Any = AorB | CorD

    assert tyro.cli(
        typ, args=["group-ab", "command-a", "--x", "1", "--y", "hello"]
    ) == CommandA(1, "hello")
    assert tyro.cli(
        typ, args=["group-ab", "command-b", "--a", "2.5", "--b", "True"]
    ) == CommandB(2.5, True)
    assert tyro.cli(
        typ, args=["group-cd", "command-c", "--p", "1", "--q", "k", "3"]
    ) == CommandC([1], {"k": 3})
    assert tyro.cli(typ, args=["group-cd", "command-d", "--value", "test"]) == CommandD(
        "test"
    )


def test_helptext_for_nested_subparsers() -> None:
    """Test that help text is generated correctly for nested structures."""
    typ: Any = (
        CommandA
        | Annotated[
            CommandB | CommandC,
            tyro.conf.subcommand(name="bc-commands", description="Commands B and C"),
        ]
    )

    # Test that we can get help without errors.
    with pytest.raises(SystemExit):
        tyro.cli(typ, args=["--help"])

    with pytest.raises(SystemExit):
        tyro.cli(typ, args=["bc-commands", "--help"])

    with pytest.raises(SystemExit):
        tyro.cli(typ, args=["bc-commands", "command-b", "--help"])


def test_defaults_in_nested_subparsers() -> None:
    """Test that defaults work correctly in nested subparser structures."""
    typ: Any = (
        CommandA | Annotated[CommandB | CommandC, tyro.conf.subcommand(name="bc-group")]
    )

    # With default, bc-group should be optional.
    # This might need adjustment based on how defaults are handled.
    result = tyro.cli(typ, args=[], default=CommandB(1.0, False))
    assert result == CommandB(1.0, False)

    # Can still override the default.
    assert tyro.cli(
        typ, args=["command-a", "--x", "1", "--y", "hi"], default=CommandB(1.0, False)
    ) == CommandA(1, "hi")
    assert tyro.cli(
        typ,
        args=["bc-group", "command-c", "--p", "1", "--q", "k", "2"],
        default=CommandB(1.0, False),
    ) == CommandC([1], {"k": 2})


def test_config_default_none_preserved() -> None:
    """A subcommand config default of None should propagate to the parsed value."""
    typ: Any = CommandA | None

    # No arguments selects the configured default of None.
    assert tyro.cli(typ, args=[], default=None) is None

    # Explicitly selecting CommandA still works.
    assert tyro.cli(
        typ, args=["command-a", "--x", "1", "--y", "hi"], default=None
    ) == CommandA(1, "hi")


def test_single_union_with_name() -> None:
    """A single-element union with a name creates an alias (no extra subparser level).

    Union[CommandB] is effectively the same as CommandB, so wrapping it in a named
    union just creates an alias for the subcommand rather than nesting."""
    typ: Any = (
        CommandA | Annotated[Union[CommandB], tyro.conf.subcommand(name="b-only")]
    )

    assert tyro.cli(typ, args=["command-a", "--x", "1", "--y", "hello"]) == CommandA(
        1, "hello"
    )
    # Single-element union flattens: b-only directly takes CommandB's arguments.
    assert tyro.cli(typ, args=["b-only", "--a", "2.5", "--b", "True"]) == CommandB(
        2.5, True
    )


def test_annotations_propagate_through_flattening() -> None:
    """Annotations on the Annotated wrapper should propagate to flattened children."""
    # Suppress marker on the wrapper should suppress all children.
    typ: Any = CommandA | Annotated[CommandB | CommandC, tyro.conf.Suppress]

    with pytest.raises(SystemExit):
        tyro.cli(typ, args=["--help"])

    # Only CommandA should be available; CommandB and CommandC are suppressed.
    assert tyro.cli(typ, args=["command-a", "--x", "1", "--y", "hello"]) == CommandA(
        1, "hello"
    )
    with pytest.raises(SystemExit):
        tyro.cli(typ, args=["command-b", "--a", "2.5", "--b", "True"])
    with pytest.raises(SystemExit):
        tyro.cli(typ, args=["command-c", "--p", "1", "--q", "k", "3"])



from typing import NewType, Tuple

import pydantic
from pydantic import BaseModel
from typing_extensions import Annotated

import tyro

Microliter = NewType("Microliter", int)


class Measurements(BaseModel):
    single: Microliter = pydantic.Field(Microliter(10))
    renamed_single: Annotated[Microliter, tyro.conf.arg(name="other_single")] = (
        pydantic.Field(Microliter(10))
    )
    pair: Tuple[Microliter, Microliter] = pydantic.Field(
        (Microliter(20), Microliter(30))
    )


IncorrectMeasurements = NewType("IncorrectMeasurements", Measurements)


def test_pydantic_with_newtype():
    assert tyro.cli(
        IncorrectMeasurements, args="--single 1 --pair 2 3".split(" ")
    ) == Measurements(
        single=Microliter(1),
        renamed_single=Microliter(10),
        pair=(Microliter(2), Microliter(3)),
    )
    assert tyro.cli(
        IncorrectMeasurements, args="--single 1 --other-single 5".split(" ")
    ) == Measurements(
        single=Microliter(1),
        renamed_single=Microliter(5),
        pair=(Microliter(20), Microliter(30)),
    )



# mypy: ignore-errors
"""Tests for self-referential (recursive) dataclass types."""

from __future__ import annotations

import contextlib
import dataclasses
import io

import pytest

import tyro
import tyro._strings


@dataclasses.dataclass(frozen=True)
class InferenceConfig:
    name: str = "Hello"
    secondary: InferenceConfig | None = None


@dataclasses.dataclass(frozen=True)
class Node:
    value: int = 0
    left: Node | None = None
    right: Node | None = None


def _skip_if_argparse(backend: str) -> None:
    """The argparse backend hits infinite recursion on self-referential types."""
    if backend == "argparse":
        pytest.skip("recursive structs are tyro-backend only")


def _get_helptext(
    f,
    args: list[str] = ["--help"],
    config: tuple = (),
) -> str:
    """Get helptext without running completion checks (which also recurse infinitely)."""
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(f, args=args, config=config)
    return tyro._strings.strip_ansi_sequences(target.getvalue())


# --- Basic recursive struct tests ---


def test_recursive_default(backend: str):
    _skip_if_argparse(backend)
    assert tyro.cli(InferenceConfig, args=[]) == InferenceConfig(
        name="Hello", secondary=None
    )


def test_recursive_set_name(backend: str):
    _skip_if_argparse(backend)
    assert tyro.cli(InferenceConfig, args=["--name", "World"]) == InferenceConfig(
        name="World", secondary=None
    )


def test_recursive_one_level(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[
            "--name",
            "Outer",
            "secondary:inference-config",
            "--secondary.name",
            "Inner",
        ],
    )
    assert result == InferenceConfig(
        name="Outer",
        secondary=InferenceConfig(name="Inner", secondary=None),
    )


def test_recursive_one_level_default_secondary_name(backend: str):
    """Nested level should use its own defaults when not overridden."""
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=["--name", "Outer", "secondary:inference-config"],
    )
    assert result == InferenceConfig(
        name="Outer",
        secondary=InferenceConfig(name="Hello", secondary=None),
    )


def test_recursive_two_levels(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[
            "--name",
            "L0",
            "secondary:inference-config",
            "--secondary.name",
            "L1",
            "secondary.secondary:inference-config",
            "--secondary.secondary.name",
            "L2",
        ],
    )
    assert result == InferenceConfig(
        name="L0",
        secondary=InferenceConfig(
            name="L1",
            secondary=InferenceConfig(name="L2", secondary=None),
        ),
    )


def test_recursive_three_levels(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[
            "--name",
            "L0",
            "secondary:inference-config",
            "--secondary.name",
            "L1",
            "secondary.secondary:inference-config",
            "--secondary.secondary.name",
            "L2",
            "secondary.secondary.secondary:inference-config",
            "--secondary.secondary.secondary.name",
            "L3",
        ],
    )
    assert result == InferenceConfig(
        name="L0",
        secondary=InferenceConfig(
            name="L1",
            secondary=InferenceConfig(
                name="L2",
                secondary=InferenceConfig(name="L3", secondary=None),
            ),
        ),
    )


def test_recursive_four_levels(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[
            "--name",
            "L0",
            "secondary:inference-config",
            "--secondary.name",
            "L1",
            "secondary.secondary:inference-config",
            "--secondary.secondary.name",
            "L2",
            "secondary.secondary.secondary:inference-config",
            "--secondary.secondary.secondary.name",
            "L3",
            "secondary.secondary.secondary.secondary:inference-config",
            "--secondary.secondary.secondary.secondary.name",
            "L4",
        ],
    )
    assert result == InferenceConfig(
        name="L0",
        secondary=InferenceConfig(
            name="L1",
            secondary=InferenceConfig(
                name="L2",
                secondary=InferenceConfig(
                    name="L3",
                    secondary=InferenceConfig(name="L4", secondary=None),
                ),
            ),
        ),
    )


def test_recursive_four_levels_terminate_with_none(backend: str):
    """Four levels deep, explicitly terminating the chain with :none."""
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[
            "secondary:inference-config",
            "secondary.secondary:inference-config",
            "secondary.secondary.secondary:inference-config",
            "secondary.secondary.secondary.secondary:none",
        ],
    )
    assert result == InferenceConfig(
        name="Hello",
        secondary=InferenceConfig(
            name="Hello",
            secondary=InferenceConfig(
                name="Hello",
                secondary=InferenceConfig(name="Hello", secondary=None),
            ),
        ),
    )


def test_recursive_select_none(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=["secondary:none"],
    )
    assert result == InferenceConfig(name="Hello", secondary=None)


def test_recursive_helptext(backend: str):
    _skip_if_argparse(backend)
    helptext = _get_helptext(InferenceConfig)
    assert "secondary:inference-config" in helptext
    assert "secondary:none" in helptext
    assert "--name" in helptext


def test_recursive_invalid_args(backend: str):
    _skip_if_argparse(backend)
    with pytest.raises(SystemExit):
        tyro.cli(InferenceConfig, args=["--nonexistent", "value"])


# --- Binary tree (multiple recursive fields) ---


def test_recursive_binary_tree_default(backend: str):
    _skip_if_argparse(backend)
    assert tyro.cli(Node, args=[]) == Node(value=0, left=None, right=None)


def test_recursive_binary_tree_one_child(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        Node,
        args=[
            "--value",
            "1",
            "left:node",
            "--left.value",
            "2",
        ],
    )
    assert result == Node(
        value=1,
        left=Node(value=2, left=None, right=None),
        right=None,
    )


def test_recursive_binary_tree_both_children(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        Node,
        args=[
            "--value",
            "1",
            "left:node",
            "--left.value",
            "2",
            "right:node",
            "--right.value",
            "3",
        ],
    )
    assert result == Node(
        value=1,
        left=Node(value=2, left=None, right=None),
        right=Node(value=3, left=None, right=None),
    )


def test_recursive_binary_tree_nested(backend: str):
    """Tree with a grandchild node."""
    _skip_if_argparse(backend)
    result = tyro.cli(
        Node,
        args=[
            "--value",
            "1",
            "left:node",
            "--left.value",
            "2",
            "left.left:node",
            "--left.left.value",
            "4",
        ],
    )
    assert result == Node(
        value=1,
        left=Node(
            value=2,
            left=Node(value=4, left=None, right=None),
            right=None,
        ),
        right=None,
    )


def test_recursive_binary_tree_helptext(backend: str):
    _skip_if_argparse(backend)
    helptext = _get_helptext(Node)
    assert "left:node" in helptext
    assert "left:none" in helptext
    assert "right:node" in helptext
    assert "right:none" in helptext


# --- Default instances ---


def test_recursive_with_default_instance(backend: str):
    _skip_if_argparse(backend)
    default = InferenceConfig(
        name="DefaultOuter",
        secondary=InferenceConfig(name="DefaultInner", secondary=None),
    )
    result = tyro.cli(InferenceConfig, args=[], default=default)
    assert result == default


def test_recursive_with_default_instance_override(backend: str):
    _skip_if_argparse(backend)
    default = InferenceConfig(
        name="DefaultOuter",
        secondary=InferenceConfig(name="DefaultInner", secondary=None),
    )
    result = tyro.cli(
        InferenceConfig,
        args=["--name", "Overridden"],
        default=default,
    )
    assert result.name == "Overridden"
    assert result.secondary == InferenceConfig(name="DefaultInner", secondary=None)


def test_recursive_with_deep_default_instance(backend: str):
    """Default instance with 3 levels of nesting should round-trip."""
    _skip_if_argparse(backend)
    default = InferenceConfig(
        name="L0",
        secondary=InferenceConfig(
            name="L1",
            secondary=InferenceConfig(name="L2", secondary=None),
        ),
    )
    result = tyro.cli(InferenceConfig, args=[], default=default)
    assert result == default


# --- Tests with CascadeSubcommandArgs ---


def test_recursive_cascade_default(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )
    assert result == InferenceConfig(name="Hello", secondary=None)


def test_recursive_cascade_one_level(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[
            "--name",
            "Outer",
            "secondary:inference-config",
            "--secondary.name",
            "Inner",
        ],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )
    assert result == InferenceConfig(
        name="Outer",
        secondary=InferenceConfig(name="Inner", secondary=None),
    )


def test_recursive_cascade_one_level_inner_default(backend: str):
    """With CascadeSubcommandArgs, the inner level should use its own default."""
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=["--name", "Outer", "secondary:inference-config"],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )
    assert result == InferenceConfig(
        name="Outer",
        secondary=InferenceConfig(name="Hello", secondary=None),
    )


def test_recursive_cascade_two_levels(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[
            "--name",
            "L0",
            "secondary:inference-config",
            "--secondary.name",
            "L1",
            "secondary.secondary:inference-config",
            "--secondary.secondary.name",
            "L2",
        ],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )
    assert result == InferenceConfig(
        name="L0",
        secondary=InferenceConfig(
            name="L1",
            secondary=InferenceConfig(name="L2", secondary=None),
        ),
    )


def test_recursive_cascade_three_levels(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[
            "--name",
            "L0",
            "secondary:inference-config",
            "--secondary.name",
            "L1",
            "secondary.secondary:inference-config",
            "--secondary.secondary.name",
            "L2",
            "secondary.secondary.secondary:inference-config",
            "--secondary.secondary.secondary.name",
            "L3",
        ],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )
    assert result == InferenceConfig(
        name="L0",
        secondary=InferenceConfig(
            name="L1",
            secondary=InferenceConfig(
                name="L2",
                secondary=InferenceConfig(name="L3", secondary=None),
            ),
        ),
    )


def test_recursive_cascade_four_levels(backend: str):
    _skip_if_argparse(backend)
    result = tyro.cli(
        InferenceConfig,
        args=[
            "--name",
            "L0",
            "secondary:inference-config",
            "--secondary.name",
            "L1",
            "secondary.secondary:inference-config",
            "--secondary.secondary.name",
            "L2",
            "secondary.secondary.secondary:inference-config",
            "--secondary.secondary.secondary.name",
            "L3",
            "secondary.secondary.secondary.secondary:inference-config",
            "--secondary.secondary.secondary.secondary.name",
            "L4",
        ],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )
    assert result == InferenceConfig(
        name="L0",
        secondary=InferenceConfig(
            name="L1",
            secondary=InferenceConfig(
                name="L2",
                secondary=InferenceConfig(
                    name="L3",
                    secondary=InferenceConfig(name="L4", secondary=None),
                ),
            ),
        ),
    )


def test_recursive_cascade_helptext(backend: str):
    _skip_if_argparse(backend)
    helptext = _get_helptext(
        InferenceConfig,
        config=(tyro.conf.CascadeSubcommandArgs,),
    )
    assert "secondary:inference-config" in helptext
    assert "secondary:none" in helptext
    assert "--name" in helptext



"""Mutually exclusive groups

:func:`tyro.conf.create_mutex_group()` can be used to create mutually exclusive
argument groups, where either exactly one (required=True) or at most one
(required=False) argument from the group can be specified.

The ``title`` parameter can be used to customize the group title in the help text.

Usage:

    python ./14_mutex.py
    python ./14_mutex.py --help
    python ./14_mutex.py --target-stream stdout
    python ./14_mutex.py --target-file /tmp/output.txt
    python ./14_mutex.py --target-stream stdout --verbose
    python ./14_mutex.py --target-file /tmp/output.txt --very-verbose
    python ./14_mutex.py --target-stream stdout --target-file /tmp/output.txt
    python ./14_mutex.py --target-stream stdout --verbose --very-verbose
"""

from pathlib import Path
from typing import Annotated, Literal

import tyro

RequiredGroup = tyro.conf.create_mutex_group(required=True, title="output target")
OptionalGroup = tyro.conf.create_mutex_group(required=False, title="verbosity level")


def main(
    # Exactly one of --target-stream or --target-file must be specified.
    target_stream: Annotated[Literal["stdout", "stderr"] | None, RequiredGroup] = None,
    target_file: Annotated[Path | None, RequiredGroup] = None,
    # Either --verbose or --very-verbose can be specified (but not both).
    verbose: Annotated[bool, OptionalGroup] = False,
    very_verbose: Annotated[bool, OptionalGroup] = False,
) -> None:
    """Demonstrate mutually exclusive argument groups."""
    if very_verbose or verbose:
        print(f"{target_stream=} {target_file=}")
    if very_verbose:
        print(f"{target_stream=} {target_file=}")


if __name__ == "__main__":
    tyro.cli(
        main,
        # `DisallowNone` prevents `None` from being a valid choice for
        # `--target-stream` and `--target-file`.
        #
        # `FlagCreatePairsOff` prevents `--no-verbose` and `--no-very-verbose` from
        # being created.
        config=(tyro.conf.DisallowNone, tyro.conf.FlagCreatePairsOff),
    )



"""Nesting in containers

Structures can be nested inside of standard containers.

.. warning::

    When placing structures inside of containers like lists or tuples, the
    length of the container must be inferrable from the annotation or default
    value.


Usage:

    python ./03_nesting_containers.py --help
"""

import dataclasses

import tyro


@dataclasses.dataclass
class RGB:
    r: int
    g: int
    b: int


@dataclasses.dataclass
class Args:
    color_tuple: tuple[RGB, RGB]
    color_dict: dict[str, RGB] = dataclasses.field(
        # We can't use mutable values as defaults directly.
        default_factory=lambda: {
            "red": RGB(255, 0, 0),
            "green": RGB(0, 255, 0),
            "blue": RGB(0, 0, 255),
        }
    )


if __name__ == "__main__":
    args = tyro.cli(Args)
    print(args)
